import logging

from django.db import models

from ...bases.listing.models import ListingItem
from ...bases.metadata.models import BaseListingItemRelation

logger = logging.getLogger(__name__)


class Availability(models.Model):
    app = models.CharField(max_length=50)
    identifier = models.CharField(max_length=250)

    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = (("app", "identifier"),)

    @property
    def metadata_name(self):
        return "available"


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(Availability, on_delete=models.CASCADE)


class ListingItemAvailability(models.Model):
    app = models.CharField(max_length=50)
    identifier = models.CharField(max_length=250)

    listingitem = models.ForeignKey(ListingItem, on_delete=models.CASCADE)

    class Meta:
        indexes = [models.Index(fields=["app", "identifier"])]



import fnmatch
import logging
from abc import abstractmethod

from unplugged import PluginBase, Schema, fields

logger = logging.getLogger(__name__)


class Notification:
    def __init__(
        self,
        topic,
        notification_type,
        title,
        body,
        permission_type=None,
        permission_config=None,
    ):
        self.topic = topic

        self.type = notification_type
        self.title = title
        self.body = body

        self.permission_type = permission_type
        self.permission_config = permission_config

    def is_allowed(self, user):
        if self.permission_type == "user":
            return user.username in self.permission_config.get("usernames", [])
        elif self.permission_type == "permission":
            return user.has_perm(self.permission_config.get("permission"))
        elif self.permission_type == "is_admin":
            return user.is_superuser
        elif self.permission_type == None:
            return True
        else:
            logger.warning(f"Unknown permission_type: {self.permission_type}")
            return False

    def match_patterns(self, patterns):
        return any(fnmatch.fnmatch(self.topic, pattern) for pattern in patterns)

    def copy(
        self,
        topic=None,
        notification_type=None,
        title=None,
        body=None,
        permission_type=None,
        permission_config=None,
    ):
        if topic is None:
            topic = self.topic

        if notification_type is None:
            notification_type = self.type

        if title is None:
            title = self.title

        if body is None:
            body = self.body

        if permission_type is None:
            permission_type = self.permission_type

        if permission_config is None:
            permission_config = self.permission_config

        return self.__class__(
            topic, notification_type, title, body, permission_type, permission_config
        )


class NotifierPlugin(PluginBase):
    plugin_type = "notifier"

    @abstractmethod
    def notify(self, notification):
        """
        Send a notification to whoever is concerned.
        """
        raise NotImplementedError



from django.contrib.auth.models import User
from django.db import models
from jsonfield import JSONField


class DefaultSetting(models.Model):
    SYSTEM_DEFAULT = "system"
    CONFIGURED_DEFAULT = "configured"
    DEFAULT_TYPE_CHOICES = (
        (SYSTEM_DEFAULT, "System Default"),
        (CONFIGURED_DEFAULT, "Configured Default"),
    )
    default_type = models.CharField(max_length=30, choices=DEFAULT_TYPE_CHOICES)

    namespace = models.CharField(max_length=100)
    key = models.CharField(max_length=100)
    value = JSONField()

    class Meta:
        unique_together = (("default_type", "namespace", "key"),)


class Setting(models.Model):
    user = models.ForeignKey(User, on_delete=models.deletion.CASCADE)
    namespace = models.CharField(max_length=100)
    key = models.CharField(max_length=100)
    value = JSONField()

    class Meta:
        unique_together = (("user", "namespace", "key"),)



import logging

from django.conf.urls import url
from unplugged import DefaultPermission, RelatedPluginField, Schema, command, fields

from ...bases.listing.handler import BaseListingService
from ...bases.listing.models import ListingItem
from ...bases.listing.schema import BaseSchema, ListingSchema
from ...plugins import InputPlugin, InputPluginManager
from .views import SectionsListView

logger = logging.getLogger(__name__)


class SectionInputSchema(Schema):
    path = fields.String(default="")
    preference = fields.Integer(default=10)
    input = RelatedPluginField(plugin_type=InputPlugin, required=True)


class SectionSchema(ListingSchema):
    inputs = fields.Nested(SectionInputSchema, many=True, default=list)


class SectionServiceSchema(BaseSchema):
    sections = fields.Nested(SectionSchema, many=True, default=list)


class SectionsService(BaseListingService):
    plugin_name = "sections"
    config_schema = SectionServiceSchema
    default_permission = DefaultPermission.ALLOW

    simpleadmin_templates = [
        {
            "template": {
                "display_name": {"simpleAdminMethod": "userInput", "required": True},
                "inputs": {
                    "simpleAdminMethod": "userInput",
                    "hideFields": ["preference"],
                },
                "name": {"simpleAdminMethod": "slugify", "source": "display_name"},
                "histories": [
                    {
                        "simpleAdminMethod": "lookupPlugin",
                        "plugin_type": "history",
                        "name": "history",
                    }
                ],
                "primary_metadata_plugin": {
                    "simpleAdminMethod": "injectablePlugin",
                    "plugin_type": "metadatahandler",
                    "required": False,
                    "traits": ["primary_metadata", "metadata_movie"],
                },
                "rebuild_automatically": True,
                "levels": [
                    {
                        "indexer": {
                            "simpleAdminMethod": "createPlugin",
                            "plugin_type": "indexer",
                            "plugin_name": "whoosh",
                            "name": "index_automanaged_%(name)s",
                            "config": {"path": "index_automanaged_%(name)s"},
                        },
                        "depth": 0,
                        "content_type": "movies",
                        "default_ordering": "-metadata_firstseen__datetime,-datetime",
                        "listing_depth": 0,
                        "tags": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "tag",
                                "name": "tag",
                            }
                        ],
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "injectedPlugin",
                                "id": "primary_metadata_plugin",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "historymetadata",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "iteminfo",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "firstseen",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "name",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "tag",
                            },
                        ],
                    }
                ],
            },
            "description": "A folder with movies in it",
            "id": "movies",
            "update_method": "modify_key",
            "modify_key": "sections",
            "display_name": "Movies",
        },
        {
            "template": {
                "display_name": {"simpleAdminMethod": "userInput", "required": True},
                "inputs": {
                    "simpleAdminMethod": "userInput",
                    "hideFields": ["preference"],
                },
                "name": {"simpleAdminMethod": "slugify", "source": "display_name"},
                "histories": [
                    {
                        "simpleAdminMethod": "lookupPlugin",
                        "plugin_type": "history",
                        "name": "history",
                    }
                ],
                "primary_metadata_plugin": {
                    "simpleAdminMethod": "injectablePlugin",
                    "plugin_type": "metadatahandler",
                    "traits": ["primary_metadata", "metadata_tv"],
                },
                "rebuild_automatically": True,
                "levels": [
                    {
                        "indexer": {
                            "simpleAdminMethod": "createPlugin",
                            "plugin_type": "indexer",
                            "plugin_name": "whoosh",
                            "name": "index_automanaged_%(name)s",
                            "config": {"path": "index_automanaged_%(name)s"},
                        },
                        "depth": 0,
                        "content_type": "tvshows",
                        "default_ordering": "-datetime",
                        "listing_depth": 0,
                        "tags": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "tag",
                                "name": "tag",
                            }
                        ],
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "injectedPlugin",
                                "id": "primary_metadata_plugin",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "tag",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "firstseen",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "name",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "tag",
                            },
                        ],
                    },
                    {
                        "indexer": None,
                        "depth": 1,
                        "content_type": "episodes",
                        "default_ordering": "-metadata_iteminfo__season,-metadata_iteminfo__episode",
                        "listing_depth": 0,
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "historymetadata",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "iteminfo",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "firstseen",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "name",
                            },
                        ],
                    },
                ],
            },
            "description": "Folder with TV shows.\n\n"
            "The folder must have the organization /tv/show/episode.\n"
            "An example could be /TVFolder/A.Good.Show/A.Good.Show.S03E03",
            "id": "tv-episode",
            "update_method": "modify_key",
            "modify_key": "sections",
            "display_name": "TV /tv/show/episode",
        },
        {
            "template": {
                "display_name": {"simpleAdminMethod": "userInput", "required": True},
                "inputs": {
                    "simpleAdminMethod": "userInput",
                    "hideFields": ["preference"],
                },
                "name": {"simpleAdminMethod": "slugify", "source": "display_name"},
                "histories": [
                    {
                        "simpleAdminMethod": "lookupPlugin",
                        "plugin_type": "history",
                        "name": "history",
                    }
                ],
                "primary_metadata_plugin": {
                    "simpleAdminMethod": "injectablePlugin",
                    "plugin_type": "metadatahandler",
                    "traits": ["primary_metadata", "metadata_tv"],
                },
                "rebuild_automatically": True,
                "levels": [
                    {
                        "indexer": {
                            "simpleAdminMethod": "createPlugin",
                            "plugin_type": "indexer",
                            "plugin_name": "whoosh",
                            "name": "index_automanaged_%(name)s",
                            "config": {"path": "index_automanaged_%(name)s"},
                        },
                        "depth": 0,
                        "content_type": "tvshows",
                        "default_ordering": "-datetime",
                        "listing_depth": 0,
                        "tags": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "tag",
                                "name": "tag",
                            }
                        ],
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "injectedPlugin",
                                "id": "primary_metadata_plugin",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "tag",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "firstseen",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "name",
                            },
                        ],
                    },
                    {
                        "indexer": None,
                        "depth": 1,
                        "content_type": "seasons",
                        "default_ordering": "-metadata_iteminfo__season",
                        "listing_depth": 0,
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "iteminfo",
                            }
                        ],
                    },
                    {
                        "indexer": None,
                        "depth": 2,
                        "content_type": "episodes",
                        "default_ordering": "-metadata_iteminfo__season,-metadata_iteminfo__episode",
                        "listing_depth": 0,
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "historymetadata",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "iteminfo",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "firstseen",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "name",
                            },
                        ],
                    },
                ],
            },
            "description": "Folder with TV shows.\n\n"
            "The folder must have the organization /tv/show/season/episode.\n"
            "An example could be /TVFolder/A.Good.Show/Season.03/A.Good.Show.S03E03",
            "id": "tv-season-episode",
            "update_method": "modify_key",
            "modify_key": "sections",
            "display_name": "TV /tv/show/season/episode",
        },
    ]

    def get_urls(self):
        return [
            url(
                r"^(?P<path>.*)$",
                SectionsListView.as_view(
                    service=self, listing_builder=self.listing_builder
                ),
            )
        ]

    def get_item(self, config, path):
        if not config.get("inputs"):
            return None

        plugin_path_pairs = []
        for input_config in config["inputs"]:
            input_path = input_config.get("path", "").strip("/")
            listing_path = f"{input_path}/{config['path']}"
            plugin_path_pairs.append((input_config["input"], listing_path.strip("/")))

        logger.info(f"Trying to create listing for {plugin_path_pairs!r} - path:{path}")
        item = InputPluginManager.get_item_multiple(plugin_path_pairs)
        if "/" not in path and "display_name" in config:
            item.id = config["display_name"]
        return item



# Generated by Django 2.2.8 on 2019-12-13 16:29

import jsonfield.fields
import picklefield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="WebSession",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app", models.CharField(max_length=30, unique=True)),
                (
                    "requests_cookies",
                    picklefield.fields.PickledObjectField(editable=False),
                ),
                ("last_modified", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="TorrentFile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app", models.CharField(max_length=30)),
                ("url", models.CharField(max_length=5000)),
                ("url_hash", models.CharField(max_length=50)),
                ("torrent_data", models.BinaryField()),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
            options={"unique_together": {("app", "url_hash")},},
        ),
        migrations.CreateModel(
            name="ListingCache",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app", models.CharField(max_length=30)),
                ("search_token", models.CharField(max_length=50)),
                ("path", models.CharField(default="", max_length=500)),
                ("listing", jsonfield.fields.JSONField()),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
            options={"unique_together": {("app", "search_token", "path")},},
        ),
        migrations.CreateModel(
            name="AvailableMedia",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app", models.CharField(max_length=30)),
                ("identifier", models.CharField(max_length=500)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
            ],
            options={"unique_together": {("app", "identifier")},},
        ),
    ]



import logging

from django.db import models

from ...bases.metadata.models import BaseListingItemRelation, BaseMetadata

logger = logging.getLogger(__name__)


class Genre(models.Model):
    name = models.CharField(max_length=50, unique=True)
    identifier = models.IntegerField(unique=True)

    def __str__(self):
        return self.name


class Season(models.Model):
    SEASON_SUMMER = "summer"
    SEASON_WINTER = "winter"
    SEASON_SPRING = "spring"
    SEASON_FALL = "fall"
    SEASON_CHOICES = (
        (SEASON_SUMMER, "Summer"),
        (SEASON_WINTER, "Winter"),
        (SEASON_SPRING, "Spring"),
        (SEASON_FALL, "Fall"),
    )

    year = models.IntegerField()
    season = models.CharField(max_length=20, choices=SEASON_CHOICES)

    class Meta:
        unique_together = (("year", "season"),)


class TVRating(models.Model):
    name = models.CharField(max_length=100, unique=True)

    def __str__(self):
        return self.name


class Type(models.Model):
    name = models.CharField(max_length=100, unique=True)

    def __str__(self):
        return self.name


class Metadata(BaseMetadata):
    genres = models.ManyToManyField(Genre)

    rating = models.DecimalField(null=True, decimal_places=2, max_digits=3)
    votes = models.IntegerField(null=True)
    duration = models.IntegerField(null=True)
    episodes = models.IntegerField(null=True)

    season = models.ForeignKey(Season, null=True, on_delete=models.deletion.CASCADE)
    tv_rating = models.ForeignKey(
        TVRating, null=True, on_delete=models.deletion.CASCADE
    )
    anime_type = models.ForeignKey(Type, null=True, on_delete=models.deletion.CASCADE)

    aired_start = models.DateField(null=True)
    aired_end = models.DateField(null=True)

    plot = models.TextField(null=True)

    @property
    def metadata_name(self):
        return "mal"

    def populate(self, config):
        from malparser import MAL
        from malparser.anime import Anime

        image_cache = config["image_cache"]

        mal = MAL()
        anime = mal.get_anime(self.identifier)
        anime.fetch()

        # cleanup if it is an update
        self.genres.clear()
        AlternativeTitle.objects.filter(metadata=self).delete()
        Related.objects.filter(from_metadata=self).delete()

        # populate
        self.title = anime.title
        self.rating = anime.statistics["Score"]
        self.votes = anime.statistics["Votes"]
        self.duration = anime.info["Duration"]
        self.episodes = anime.info["Episodes"]
        self.aired_start = anime.aired.get("Aired_start", None)
        self.aired_end = anime.aired.get("Aired_end", None)
        self.plot = anime.synopsis

        season = anime.aired.get("Season", None)
        if season:
            season, created = Season.objects.get_or_create(
                season=season[0].lower(), year=season[1]
            )
        self.season = season

        tv_rating, created = TVRating.objects.get_or_create(name=anime.info["Rating"])
        self.tv_rating = tv_rating

        anime_type, created = Type.objects.get_or_create(name=anime.info["Type"])
        self.anime_type = anime_type

        for g in anime.info["Genres"]:
            genre, created = Genre.objects.get_or_create(
                identifier=g["id"], defaults={"name": g["name"]}
            )
            if not created and genre.name != g["name"]:
                genre.name = g["name"]
                genre.save()
            self.genres.add(genre)

        self.cover = anime.cover
        image_cache.get_image_path(self.cover)

        for language, titles in anime.alternative_titles.items():
            for title in titles:
                AlternativeTitle.objects.create(
                    title=title, language=language, metadata=self
                )

        for relation_type, related in anime.related.items():
            for item in related:
                if not isinstance(item, Anime):
                    continue

                try:
                    to_metadata = Metadata.objects.get(identifier=item.mal_id)
                except Metadata.DoesNotExist:
                    to_metadata = None

                Related.objects.create(
                    from_metadata=self,
                    relation_type=relation_type,
                    to_metadata=to_metadata,
                    to_metadata_identifier=item.mal_id,
                )


class AlternativeTitle(models.Model):
    title = models.CharField(max_length=500)
    language = models.CharField(max_length=100)
    metadata = models.ForeignKey(Metadata, on_delete=models.deletion.CASCADE)


class Related(models.Model):
    from_metadata = models.ForeignKey(
        Metadata, related_name="related", on_delete=models.deletion.CASCADE
    )
    relation_type = models.CharField(max_length=50)
    to_metadata = models.ForeignKey(
        Metadata,
        null=True,
        related_name="reverse_related",
        on_delete=models.deletion.CASCADE,
    )
    to_metadata_identifier = models.CharField(max_length=100)


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(
        Metadata, db_index=True, on_delete=models.deletion.CASCADE
    )



from django.conf import settings
from unplugged import RelatedPluginField, Schema, ServicePlugin, fields

from ...plugins import HistoryPlugin, IndexerPlugin, MetadataHandlerPlugin, TagPlugin


class ListingLevelSchema(Schema):
    depth = fields.Integer(required=True)
    listing_depth = fields.Integer(required=True)
    metadata_handlers = fields.List(
        RelatedPluginField(plugin_type=MetadataHandlerPlugin), many=True, default=list
    )
    indexer = RelatedPluginField(plugin_type=IndexerPlugin, required=False)
    content_type = fields.String(
        enum=settings.TRIDENTSTREAM_CONTENT_TYPES,
        enum_names=settings.TRIDENTSTREAM_CONTENT_TYPES_NAMES,
        default="",
    )
    tags = fields.List(
        RelatedPluginField(plugin_type=TagPlugin), many=True, default=list
    )
    background_recheck = fields.Boolean(default=False)
    default_ordering = fields.String(default="")


class ListingSchema(Schema):
    name = fields.String(required=True)
    display_name = fields.String(default="", ui_schema={"ui:title": "Name"})
    levels = fields.Nested(ListingLevelSchema, many=True, default=list)
    histories = fields.List(
        RelatedPluginField(plugin_type=HistoryPlugin), many=True, default=list
    )
    rebuild_automatically = fields.Boolean(default=True)


class BaseSchema(Schema):
    display_name = fields.String(
        default="", description="Display Name", ui_schema={"ui:title": "Name"}
    )
    player_service = RelatedPluginField(plugin_type=ServicePlugin, traits=["player"])



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.bittorrentclients.deluge"
    verbose_name = "Deluge BittorrentClient"
    label = "bittorrentclients_deluge"

    def ready(self):
        from .handler import DelugeBittorrentClientPlugin  # NOQA



import logging
from collections import defaultdict

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Count, Field, Q
from unplugged import JSONAPIObject, command

from .models import BaseMetadataResolutionLink, BaseUpdatable

logger = logging.getLogger(__name__)


class ListingItemRelinkingMixin:
    """
    Mixin to link Metadata and ListingItems according to the metadata_mapping.

    Tries to do it intelligently.
    """

    listing_item_relation_model = None

    def listingitem_relinking(self, listing_item_root, metadata_mapping):
        logger.debug("Figuring out what we can delete or just update")
        existing_listingitem_relations = defaultdict(set)
        actual_existing_listingitem_relations = {}
        for (
            pk,
            listingitem_id,
            metadata_id,
            user_id,
        ) in self.listing_item_relation_model.objects.filter(
            Q(listingitem__parent=listing_item_root) | Q(listingitem=listing_item_root)
        ).values_list(
            "pk", "listingitem_id", "metadata_id", "user_id"
        ):
            existing_listingitem_relations[listingitem_id].add((metadata_id, user_id))
            actual_existing_listingitem_relations[
                (listingitem_id, metadata_id, user_id)
            ] = pk
        logger.debug(
            "Found %i listingitems with existing relations"
            % (len(existing_listingitem_relations),)
        )

        remove_relations = []
        create_relations = []

        listing_item_model = type(listing_item_root)
        listingitems = listing_item_model.objects.filter(
            Q(pk=listing_item_root.pk) | Q(parent=listing_item_root)
        )
        for listingitem_id in listingitems.values_list("pk", flat=True):
            metadata_ids = metadata_mapping.get(listingitem_id, set())
            existing_metadata_ids = existing_listingitem_relations.get(
                listingitem_id, set()
            )

            logger.trace(
                f"Comparing existing:{existing_metadata_ids!r} and expected:{metadata_ids!r}"
            )
            for metadata_id, user_id in existing_metadata_ids - metadata_ids:
                logger.trace(
                    f"Removing relation between listingitem_id:{listingitem_id} and metadata_id:{metadata_id} and user_id:{user_id}"
                )
                relation = actual_existing_listingitem_relations[
                    (listingitem_id, metadata_id, user_id)
                ]
                remove_relations.append(relation)

            for metadata_id, user_id in metadata_ids - existing_metadata_ids:
                logger.trace(
                    f"Creating relation between listingitem_id:{listingitem_id} and metadata_id:{metadata_id} and user_id:{user_id}"
                )
                create_relations.append(
                    self.listing_item_relation_model(
                        listingitem_id=listingitem_id,
                        metadata_id=metadata_id,
                        user_id=user_id,
                    )
                )

        logger.info(f"Creating {len(create_relations)} relations")
        self.listing_item_relation_model.objects.bulk_create(create_relations)

        logger.info(f"Deleting {len(remove_relations)} relations")
        self.listing_item_relation_model.objects.filter(
            pk__in=remove_relations
        ).delete()


class UnknownEmbedMethodException(Exception):
    pass


class StrDictProxy(dict):
    def __str__(self):
        return self.get("__str__")


class PrefetchProxy:
    def __init__(self, obj, prefetch_fields):
        self._obj = obj
        self._prefetch_fields = set(prefetch_fields)
        self._prefetch_related_denormalized = getattr(
            obj, "prefetch_related_denormalized", {}
        )

    def __setattr__(self, key, value):
        if key.startswith("_") and not key.startswith("__"):
            return super(PrefetchProxy, self).__setattr__(key, value)
        else:
            return self._obj.__setattr__(key, value)

    def __getattr__(self, key):
        if (
            self._prefetch_fields
            and key in self._prefetch_fields
            and self._prefetch_related_denormalized
            and key in self._prefetch_related_denormalized
        ):
            value = self._prefetch_related_denormalized[key]
            if isinstance(value, dict):
                return StrDictProxy(value)
            elif isinstance(value, list):
                return [StrDictProxy(v) for v in value]
            else:
                return value
        else:
            return getattr(self._obj, key)


class PopulateMetadataJSONAPIMixin:
    model = None
    listing_item_relation_model = None
    select_related = []
    prefetch_related = []
    serializer = None
    metadata_embed_method = "include"  # choices: include, relate, embed, skip
    prefetch_related_denormalized_fields = [
        "AutoField",
        "BooleanField",
        "CharField",
        "DecimalField",
        "IntegerField",
        "SmallIntegerField",
        "TextField",
    ]

    def has_prefetch_related_denormalized(self):
        return hasattr(self.model, "prefetch_related_denormalized")

    def get_relations(self, user, listingitem_ids):
        prefetch_related = ["metadata"]
        if not self.has_prefetch_related_denormalized():
            prefetch_related += [f"metadata__{x}" for x in self.prefetch_related]
        select_related = [f"metadata__{x}" for x in self.select_related]
        relations = (
            self.listing_item_relation_model.objects.filter(
                listingitem__in=listingitem_ids
            )
            .filter(Q(user__isnull=True) | Q(user=user))
            .select_related(*select_related)
            .prefetch_related(*prefetch_related)
        )

        retval = defaultdict(list)
        for relation in relations:
            retval[relation.listingitem_id].append(relation.metadata)

        return retval

    def get_jsonapi_type(self):
        return "metadata_%s" % self.plugin_name

    def populate_metadata_jsonapi(self, request, root):
        if self.metadata_embed_method == "skip":
            return

        logger.info("Populating JSONAPI root with metadata")

        obj_type = self.get_jsonapi_type()

        items = {
            obj._original_object.id: obj
            for obj in (root.data + list(root.included.values()))
            if obj._original_object and hasattr(obj._original_object, "id")
        }

        relations = self.get_relations(request.user, items.keys())

        jsonapi_obj_mapping = {}  # maps listitem to metadata jsonapi objects
        logger.info(f"Seems like we should embed the serialized metadata {self.name}")

        has_prefetch_related_denormalized = self.has_prefetch_related_denormalized()
        metadatas = [
            (
                has_prefetch_related_denormalized
                and PrefetchProxy(m, self.prefetch_related)
                or m
            )
            for m in set(v for vs in relations.values() for v in vs)
        ]

        try:
            self.model._meta.get_field("last_update_status")
            has_update_status = True
        except FieldDoesNotExist:
            has_update_status = False
        # else:
        #     metadatas = metadatas.filter(last_update_status__in=['success', 'do-not-fetch'])

        usable_metadata = []
        empty_metadata = []
        for metadata in metadatas:
            if not has_update_status or metadata.last_update_status == "success":
                usable_metadata.append(metadata)
                logger.trace(
                    f"Serializing a populated metadata entry for {metadata.identifier} / {metadata!r}"
                )
            else:
                empty_metadata.append(metadata)
                logger.trace(
                    f"Serializing an unpopulated metadata entry for {metadata.identifier} / {metadata!r} / {getattr(metadata, 'last_update_status', 'no-status')}"
                )

        all_serialized = self.serializer(
            usable_metadata,
            context={"request": request, "config": self.config},
            many=True,
        ).data + [
            {"id": metadata.identifier, "populated": False}
            for metadata in empty_metadata
        ]

        for metadata, serialized in zip(
            usable_metadata + empty_metadata, all_serialized
        ):
            identifier = f"{self.plugin_name}:{metadata.identifier}"

            if self.metadata_embed_method in ["include", "relate"]:
                metadata.__config__ = self.config
                obj = JSONAPIObject(obj_type, identifier, metadata)
                obj.update(serialized)
            elif self.metadata_embed_method == "embed":
                obj = {"metadata:%s" % self.plugin_name: serialized}
            else:
                raise UnknownEmbedMethodException(
                    f"Unknown embed method: {self.metadata_embed_method!r}"
                )

            jsonapi_obj_mapping[identifier] = obj
            logger.trace(f"Created metadata object {identifier} with metadata")

        for listingitem_id, metadatas in relations.items():
            for metadata in metadatas:
                identifier = "%s:%s" % (self.plugin_name, metadata.identifier)
                if self.metadata_embed_method == "embed":
                    m = jsonapi_obj_mapping.get(identifier, {})
                    items[listingitem_id].update(m)
                elif self.metadata_embed_method in ["include", "relate"]:
                    if identifier not in jsonapi_obj_mapping:
                        metadata.__config__ = self.config
                        obj = JSONAPIObject(
                            obj_type, identifier, metadata, populated=False
                        )
                        jsonapi_obj_mapping[identifier] = obj
                        logger.trace(
                            f"Created metadata object {identifier} without metadata"
                        )

                    local = self.metadata_embed_method == "relate"
                    logger.trace(
                        f"Added relationship between listingitem_id:{listingitem_id} metadata_identifier:{identifier}"
                    )
                    items[listingitem_id].add_relationship(
                        obj_type, jsonapi_obj_mapping[identifier], local=local
                    )
                else:
                    raise UnknownEmbedMethodException(
                        f"Unknown embed method: {self.metadata_embed_method!r}"
                    )

    def get_prefetch_related_denormalized(self, metadata):
        def serialize_obj(obj):
            result = {"__str__": obj.__str__()}
            for field in obj._meta.get_fields(False):
                if not isinstance(field, Field):
                    continue

                if (
                    field.get_internal_type()
                    not in self.prefetch_related_denormalized_fields
                ):
                    continue

                field_name = field.name
                result[field_name] = getattr(obj, field_name)

            return result

        result = {}
        for field_name in self.prefetch_related:
            result[field_name] = []
            for obj in getattr(metadata, field_name).all():
                result[field_name].append(serialize_obj(obj))

        return result

    @command(
        "refresh_denormalized_cache",
        display_name="Refresh Denormalized Cache",
        description="Clear and rebuild cache used to speed up metadata serialization",
    )
    def refresh_all_prefetch_related_denormalized(self):
        for metadata in self.model.objects.filter(last_update_status="success"):
            metadata.prefetch_related_denormalized = self.get_prefetch_related_denormalized(
                metadata
            )
            metadata.save()
            logger.trace(
                f"Updated Prefetch Related Denormalized Cache for {metadata!r}"
            )


class GetMetadataStatsMixin:
    model = None
    metadata_resolution_link_model = None

    def get_stats(self):
        stats = []

        if self.model and issubclass(self.model, BaseUpdatable):
            for result in self.model.objects.values("last_update_status").annotate(
                Count("last_update_status")
            ):
                k = f"Model update status: {result['last_update_status']}"
                stats.append({"title": k, "value": result["last_update_status__count"]})

        if self.metadata_resolution_link_model and issubclass(
            self.metadata_resolution_link_model, BaseUpdatable
        ):
            for result in self.metadata_resolution_link_model.objects.values(
                "last_update_status"
            ).annotate(Count("last_update_status")):
                k = f"Resolution update status: {result['last_update_status']}"
                stats.append({"title": k, "value": result["last_update_status__count"]})

            m = self.metadata_resolution_link_model.objects.filter(
                last_update_status="success"
            )
            stats.append(
                {
                    "title": "Resolution resolved",
                    "value": m.filter(metadata__isnull=False).count(),
                }
            )
            stats.append(
                {
                    "title": "Resolution failed",
                    "value": m.filter(metadata__isnull=True).count(),
                }
            )

        return stats


class ResetFailedMixin:
    model = None
    metadata_resolution_link_model = None

    @command(
        display_name="Reset failed fetches",
        description="Retry fetching metadata for previously failed fetches",
    )
    def command_reset_failed_fetches(self):
        if not self.model or not issubclass(self.model, BaseUpdatable):
            return

        logger.debug(f"Requeuing failed for plugin {self.plugin_name}/{self.name}")
        self.model.objects.filter(
            last_update_status=BaseUpdatable.UPDATE_STATUS_FAILED
        ).update(last_update_status=BaseUpdatable.UPDATE_STATUS_PENDING)

        if hasattr(self, "schedule_update"):
            self.schedule_update()

    @command(
        display_name="Clear failed searches",
        description="Remove failed searches so the system can try searching for them again",
    )
    def command_clear_failed_searches(self):
        if not self.metadata_resolution_link_model or not issubclass(
            self.metadata_resolution_link_model, BaseMetadataResolutionLink
        ):
            return

        search_resolve = self.config.get("search_resolve")
        if not search_resolve:
            return

        logger.debug(
            f"Wiping all failed search resolves for resolver {search_resolve} plugin {self.plugin_name}/{self.name}"
        )
        self.metadata_resolution_link_model.objects.filter(
            search_resolve=search_resolve, metadata__isnull=True
        ).exclude(last_update_status=BaseUpdatable.UPDATE_STATUS_PENDING).delete()

        if hasattr(self, "schedule_update"):
            self.schedule_update()


class CoverSerializerMixin:
    def get_cover_url(self, obj):
        if obj.cover:
            request = self.context["request"]
            image_cache = self.context["config"]["image_cache"]
            return image_cache.get_image_url(request, obj.cover)



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.metadata.firstseen"
    verbose_name = "First Seen Metadata"
    label = "metadata_firstseen"

    def ready(self):
        from .handler import FirstSeenMetadataHandlerPlugin  # NOQA



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.metadata.tag"
    verbose_name = "Tag Metadata"
    label = "metadata_tag"

    def ready(self):
        from .handler import TagMetadataHandlerPlugin  # NOQA
        from .tag import TagTagPlugin  # NOQA



from django.utils.deprecation import MiddlewareMixin


class DisableCSRFMiddleware(MiddlewareMixin):
    def process_request(self, request):
        setattr(request, "_dont_enforce_csrf_checks", True)



import logging

from rest_framework import serializers
from unplugged import Schema

from ...bases.metadata.remotemetadata import RemoteMetadataHandlerPlugin
from .models import ListingItemRelation, Metadata

logger = logging.getLogger(__name__)


class NameMetadataHandlerPlugin(RemoteMetadataHandlerPlugin):
    plugin_name = "name"
    serializer = serializers.Serializer
    model = Metadata
    listing_item_relation_model = ListingItemRelation

    config_schema = Schema
    metadata_embed_method = "skip"
    prepopulate_metadata = False
    priority = 50

    def get_identifier(self, item):
        return item.id.lower()



from abc import abstractmethod

from unplugged import PluginBase


class ConfigPlugin(PluginBase):
    plugin_type = "config"
    default_types = ["configured", "system"]

    @abstractmethod
    def get_default_config(self, namespace, key):
        """
        Returns the default configuration for a given key in a
        namespace, if it exists
        """
        raise NotImplementedError

    @abstractmethod
    def get_user_config(self, user, namespace, key):
        """
        Returns the user configuration for a given key in a given namespace
        if it exists, otherwise it returns default config.
        """
        raise NotImplementedError

    @abstractmethod
    def set_default_config(self, default_type, namespace, key, value):
        """
        Sets a default config for a given key in a namespace to value.
        """
        raise NotImplementedError

    @abstractmethod
    def set_user_config(self, user, namespace, key, value):
        """
        Sets a user config for a given key in a namespace to value.
        """
        raise NotImplementedError

    @abstractmethod
    def set_config_schema(self, namespace, key, schema):
        """
        Set a marshmallow schema usable for this specific namespace/key.

        Should not be persistent
        """
        raise NotImplementedError

    @abstractmethod
    def get_config_schema(self, namespace, key):
        """
        Get a marshmallow schema usable for this specific namespace/key if available.

        """
        raise NotImplementedError



import logging
import re

from django.utils.text import slugify
from unplugged import RelatedPluginField, Schema, fields

from ...bases.metadata.remotemetadata import (
    BaseMetadataSerializer,
    RemoteMetadataHandlerPlugin,
)
from .filters import MetadataFilter
from .models import ListingItemRelation, Metadata
from .plugins import ItemInfoPlugin

logger = logging.getLogger(__name__)


class ItemInfoTypeSchema(Schema):
    item_info = RelatedPluginField(plugin_type=ItemInfoPlugin)
    priority = fields.Integer(default=10)


class ItemInfoSchema(Schema):
    item_infos = fields.Nested(ItemInfoTypeSchema, many=True)


class ItemInfoMetadataSerializer(BaseMetadataSerializer):
    class Meta(BaseMetadataSerializer.Meta):
        model = Metadata
        exclude = BaseMetadataSerializer.Meta.exclude + ("cover", "name")


class ItemInfoMetadataHandlerPlugin(RemoteMetadataHandlerPlugin):
    plugin_name = "iteminfo"
    serializer = ItemInfoMetadataSerializer
    model = Metadata
    filter = MetadataFilter
    listing_item_relation_model = ListingItemRelation

    config_schema = ItemInfoSchema
    metadata_embed_method = "relate"
    prepopulate_metadata = True
    priority = 10

    def get_identifier(self, item):
        return slugify(re.sub(r"[-_\. ]+", "-", item.id))



from ..shelve.tests import ShelfDatabaseTest
from .handler import LMDBDatabasePlugin


class LMDBDatabaseTest(ShelfDatabaseTest):
    def open_database(self):
        return LMDBDatabasePlugin({"path": self.db_path})



import logging

logger = logging.getLogger(__name__)


class debugdict(dict):
    synced = False
    closed = False

    def sync(self):
        self.synced = True

    def close(self):
        self.closed = True



import os
from abc import abstractmethod

from django.conf import settings
from unplugged import PluginBase


class PathIndexer:
    @abstractmethod
    def index(self, path, title):
        """
        Index a path.
        """

    @abstractmethod
    def commit(self):
        """
        Commit the writes.
        """

    @abstractmethod
    def delete(self, path):
        """
        Delete an index for a path
        """


class IndexerPlugin(PluginBase):
    plugin_type = "indexer"

    @abstractmethod
    def clear(self, parent_path):
        """
        Clear the index for a given parent_path.
        """

    @abstractmethod
    def get_writer(self, parent_path):
        """
        Get a writer for parent_path.
        """

    @abstractmethod
    def search(self, path, query):
        """
        Search for query in path.
        """

    def get_database_path(self, path):
        """Turn a relative into an absolute path and make sure it exists"""
        path = path or f"{self.plugin_type}_{self.name}"
        if not os.path.isabs(path):
            path = os.path.join(settings.DATABASE_ROOT, path)

        if not os.path.isdir(path):
            os.makedirs(path)

        return path



import logging

from rest_framework import serializers

from ...bases.metadata.linkingmetadata import LinkingMetadataHandlerPlugin
from .filters import MetadataFilter
from .models import ListingItemRelation, Tag

logger = logging.getLogger(__name__)


class TagSerializer(serializers.ModelSerializer):
    id = serializers.SerializerMethodField("get_metadata_identifier")
    type = serializers.SerializerMethodField("get_metadata_type")
    plugin_name = serializers.SerializerMethodField()

    class Meta:
        model = Tag
        fields = ("id", "type", "plugin_name", "tag_name")

    def get_metadata_identifier(self, obj):
        return obj.metadata.identifier

    def get_metadata_type(self, obj):
        return "metadata_%s" % obj.metadata.metadata_name

    def get_plugin_name(self, obj):
        return "tag"


class TagMetadataHandlerPlugin(LinkingMetadataHandlerPlugin):
    plugin_name = "tag"
    priority = -10

    prefetch_related = ["metadata"]
    serializer = TagSerializer
    model = Tag
    listing_item_relation_model = ListingItemRelation
    metadata_link_model = Tag
    metadata_embed_method = "relate"
    filter = MetadataFilter
    user_field = "user"

    __traits__ = ["tag"]

    def get_metadata(self, request, identifier):
        pass



import logging

from django.contrib.auth.models import User
from unplugged import RelatedPluginField, Schema, fields
from wampyre.realm import realm_manager

from ...plugins import ConfigPlugin, NotifierPlugin

logger = logging.getLogger(__name__)


class WAMPSchema(Schema):
    config = RelatedPluginField(plugin_type=ConfigPlugin)


class NotifyFilterSchema(Schema):
    patterns = fields.List(fields.String, many=True, default=list)


class WAMPNotifierHandlerPlugin(NotifierPlugin):
    plugin_name = "wamp"
    config_schema = WAMPSchema

    default_config = {"patterns": ["admin.*"]}

    def __init__(self, config):
        self.config = config

        config_plugin = config.get("config")
        if config_plugin:
            namespace, key = self.get_config_values()
            config_plugin.set_default_config(
                "system", namespace, key, self.default_config
            )
            config_plugin.set_config_schema(namespace, key, NotifyFilterSchema)

    def get_config_values(self):
        return f"{self.plugin_type}:{self.name}", "notification_filters"

    def notify(self, notification):
        config_plugin = self.config.get("config")
        if not config_plugin:
            return

        namespace, key = self.get_config_values()

        for realm in realm_manager.get_realms():  # TODO: optimize !
            if not realm.realm.startswith("user."):
                continue

            username = realm.realm[5:]
            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                continue

            if not notification.is_allowed(user):
                continue

            config = config_plugin.get_user_config(user, namespace, key)

            if not notification.match_patterns(config.get("patterns", [])):
                continue

            realm.publish(
                {},
                "notification",
                args=(),
                kwargs={
                    "topic": notification.topic,
                    "type": notification.type,
                    "title": notification.title,
                    "body": notification.body,
                },
            )



import functools
import logging

from django.urls import path
from unplugged import Schema, ServicePlugin

from .ws import PlayerConsumer

logger = logging.getLogger(__name__)


def inject_into_scope(func, extra_scope):
    @functools.wraps(func)
    def inner(scope, *args, **kwargs):
        scope.update(extra_scope)
        return func(scope, *args, **kwargs)

    return inner


class PlayerServicePlugin(ServicePlugin):
    plugin_name = "player"
    config_schema = Schema

    __traits__ = ["player"]

    def __init__(self, config):
        self.player_coordinators = {}

    def get_channels(self):
        return [("", inject_into_scope(PlayerConsumer, {"service": self}))]

    def unload(self):
        for player_cordinator in self.player_coordinators.values():
            player_cordinator.shutdown()

        super().unload()

    def get_player_coordinator(self, user):
        if user not in self.player_coordinators:
            self.player_coordinators[user] = PlayerCoordinator(user)

        return self.player_coordinators[user]

    def play(self, payload, viewstate, player_id):
        logger.debug(f"Sending play to player_id:{player_id} user:{viewstate.user!r}")
        player_coordinator = self.get_player_coordinator(viewstate.user)

        player = player_coordinator.get_player(player_id)
        # TODO: ensure player exists

        player.connect_viewstate(viewstate)
        player.command("play", payload, viewstate.serialize())


class Player:
    def __init__(
        self, player_coordinator, websocket, player_id, name, commands, options
    ):
        self.player_coordinator = player_coordinator
        self.websocket = websocket
        self.player_id = player_id
        self.name = name
        self.state = "stopped"
        self.values = {}
        self.commands = commands
        self.options = options
        self.viewstate = None

    def command(self, method, *args):
        self.websocket.jsonrpc_method(method, *args)

    def change_state(
        self, state, values, viewstate_id=None
    ):  # we need to viewstate_id something
        if state != self.state:
            logger.debug(
                f"Main state changed for {self.player_id} from {self.state} to {state}, clearing values"
            )
            self.values = {}

        self.state = state
        self.values.update(values)
        if values and self.viewstate is not None and self.viewstate.id == viewstate_id:
            logger.debug(f"Updating viewstate {viewstate_id}")
            self.viewstate.update(values)

        self.publish_state()

    def request_state(self, state, values):
        logger.debug(f"Requesting state on id:{self.player_id} {state}/{values!r}")
        self.command("request_state", state, values)

    def connect_viewstate(self, viewstate):
        logger.debug(f"Connected player {self.player_id} with viewstate {viewstate.id}")
        self.viewstate = viewstate
        self.viewstate.player_connected = True

    def close(self):
        logger.debug(f"Closing connection to player_id:{self.player_id}")
        self.websocket.close()

    def serialize(self):
        return {
            "player_id": self.player_id,
            "name": self.name,
            "state": self.state,
            "values": self.values,
            "options": self.options,
            "commands": self.commands,
        }

    def publish_state(self):
        self.player_coordinator.command("update", self.serialize())


class PlayerCoordinator:
    def __init__(self, user):
        self.user = user
        self.players = {}
        self.controllers = []

    def add_player(self, websocket, player_id, name, commands, options):
        self.remove_player(player_id)

        player = Player(self, websocket, player_id, name, commands, options)
        self.players[player_id] = player

        player.publish_state()

    def remove_player(self, player_id):
        player = self.players.pop(player_id, None)
        if player:
            logger.debug(f"Player {player_id} gone")
            player.close()

    def get_player(self, player_id):
        return self.players.get(player_id)

    def add_controller(self, websocket):
        logger.debug("Controller added")

        for player in self.players.values():
            websocket.jsonrpc_method("update", player.serialize())

        self.controllers.append(websocket)

    def remove_controller(self, websocket):
        logger.debug("Controller gone")
        if websocket in self.controllers:
            self.controllers.remove(websocket)

    def shutdown(self):
        for player in self.players.values():
            player.close()

        for controller in self.controllers:
            controller.close()

    def command(self, method, *args):
        for controller in self.controllers:
            controller.jsonrpc_method(method, *args)



import logging

from django.contrib.contenttypes.models import ContentType
from unplugged import RelatedPluginField, Schema

from ...plugins import MetadataHandlerPlugin, TagPlugin

logger = logging.getLogger(__name__)


class TagSchema(Schema):
    tag_metadata = RelatedPluginField(plugin_type=MetadataHandlerPlugin, traits=["tag"])


class TagTagPlugin(TagPlugin):
    plugin_name = "tag"
    config_schema = TagSchema

    def __init__(self, config):
        self.tag = config["tag_metadata"]

    def tag_listingitem(self, config, user, listingitem, tag_name):
        level_config = config.get("level", {})

        for metadata_handler in level_config.get("metadata_handlers", []):
            if not metadata_handler.linkable:
                continue

            for metadata in listingitem.get_metadata(metadata_handler.plugin_name):
                content_type = ContentType.objects.get_for_model(metadata)

                try:
                    metadata_link = self.tag.metadata_link_model.objects.get(
                        user=user,
                        tag_name=tag_name,
                        object_id=metadata.pk,
                        content_type=content_type,
                    )
                except self.tag.metadata_link_model.DoesNotExist:
                    metadata_link = self.tag.metadata_link_model.objects.create(
                        user=user, tag_name=tag_name, metadata=metadata
                    )

                self.tag.listing_item_relation_model.objects.get_or_create(
                    metadata=metadata_link, listingitem=listingitem, user=user
                )

    def untag_listingitem(self, config, user, listingitem, tag_name):
        level_config = config.get("level", {})

        for metadata_handler in level_config.get("metadata_handlers", []):
            if not metadata_handler.linkable:
                continue

            for metadata in listingitem.get_metadata(metadata_handler.plugin_name):
                content_type = ContentType.objects.get_for_model(metadata)
                self.tag.metadata_link_model.objects.filter(
                    user=user,
                    tag_name=tag_name,
                    object_id=metadata.pk,
                    content_type=content_type,
                ).delete()



import logging
import os

from django.http import HttpResponse
from django.views import View
from rest_framework import permissions

logger = logging.getLogger(__name__)


class IndexView(View):
    service = None
    permission_classes = (permissions.AllowAny,)

    def get(self, request, *args, **kwargs):
        index_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "static", "index.html"
        )
        if not os.path.isfile(index_file):
            return HttpResponse(
                "No webinterface bundled with this installation, sorry",
                content_type="text/plain",
            )

        with open(index_file, "r") as f:
            index_data = f.read()

        index_data = index_data.replace(
            '<base href="/">', f'<base href="/{self.service.name}/">'
        )

        return HttpResponse(index_data, content_type="text/html")



from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from django.conf.urls import include, url
from unplugged.routing import urlpatterns

application = ProtocolTypeRouter(
    {"websocket": AuthMiddlewareStack(URLRouter(urlpatterns))}
)



from django.conf.urls import url
from unplugged import Schema, ServicePlugin

from .views import MetadataView


class MetadataServicePlugin(ServicePlugin):
    plugin_name = "metadata"
    config_schema = Schema
    __traits__ = ["metadata_server"]

    def get_urls(self):
        return [
            url(
                "^(?P<metadata_handler>[a-z_]+)/(?P<identifier>[^/]+)/?$",
                MetadataView.as_view(service=self),
                name="metadata",
            )
        ]



import logging
import time
from collections import defaultdict
from threading import Event, Lock

from django.conf import settings
from django.core.exceptions import FieldDoesNotExist
from django.db import transaction
from django.db.models import Q
from django.utils.timezone import now
from rest_framework import serializers
from unplugged import threadify

from twisted.internet import reactor

from ...plugins import MetadataHandlerPlugin
from .mixins import (
    CoverSerializerMixin,
    GetMetadataStatsMixin,
    ListingItemRelinkingMixin,
    PopulateMetadataJSONAPIMixin,
    ResetFailedMixin,
)
from .schema import MetadataSchema

logger = logging.getLogger(__name__)


class RemoteMetadataHandlerPluginMetaclass(MetadataHandlerPlugin.__class__):
    def __init__(cls, name, bases, dct):
        setattr(cls, "currently_updating_lock", Lock())
        setattr(cls, "event_queue", {})

        super(RemoteMetadataHandlerPluginMetaclass, cls).__init__(name, bases, dct)


class RemoteMetadataHandlerPlugin(
    ListingItemRelinkingMixin,
    PopulateMetadataJSONAPIMixin,
    GetMetadataStatsMixin,
    ResetFailedMixin,
    MetadataHandlerPlugin,
    metaclass=RemoteMetadataHandlerPluginMetaclass,
):
    model = None
    listing_item_relation_model = None
    currently_updating_lock = None
    serializer = None
    filter = None
    fulltext_fields = []
    linkable = True
    prepopulate_metadata = False
    metadata_resolution_link_model = None

    config_schema = MetadataSchema

    _do_unload = False
    _job_schedule = None

    def get_metadata(self, request, identifier):
        do_update = False
        try:
            metadata = self.model.objects.get(identifier=identifier)
            if metadata.last_update_status in ["do-not-fetch", "pending"]:
                metadata.priority = 50
                metadata.last_update_status = "pending"
                metadata.plugin = self._plugin_obj
                metadata.save()
                do_update = True
        except self.model.DoesNotExist:  # probably contains thread problems
            self.model(
                identifier=identifier, priority=50, plugin=self._plugin_obj
            ).save()
            do_update = True

        if do_update:
            self.schedule_update()
            if identifier not in self.event_queue:
                self.event_queue[identifier] = Event()
            e = self.event_queue[identifier]
            logger.info(
                f"Metadata {self.name}/{identifier} not found in database, waiting for it"
            )
            if not e.wait(15):
                return None

            if identifier in self.event_queue:
                del self.event_queue[identifier]

            try:
                metadata = self.model.objects.get(identifier=identifier, populated=True)
            except self.model.DoesNotExist:
                return None

        return self.serializer(
            metadata, context={"request": request, "config": self.config}
        ).data

    def get_identifier(self, item):
        trigger = f"metadata:{self.plugin_name}:id"
        return item.get(trigger)

    def link_metadata_listingitems(self, listing_build, fetch_metadata=True):
        logger.debug("Linking sections with metadata")

        listing_item_root = listing_build.listing_item_root
        item_creation_mapping = listing_build.item_creation_mapping
        listingitem_mapping = listing_build.listingitem_mapping
        index_writer = listing_build.index_writer
        existing_listing_items = listing_build.existing_listing_items
        listing_last_update = listing_build.listing_last_update

        listing_item_model = type(listing_item_root)

        try:
            self.model._meta.get_field("name")
            has_name = True
        except FieldDoesNotExist:
            has_name = False

        logger.debug("Finding all the metadata we need")
        if has_name:
            identifier_name_mapping = {}
        required_metadata = set()
        metadata_mapping = defaultdict(set)
        missing_metadata = []
        for path, item in item_creation_mapping.items():
            listingitem = listingitem_mapping.get(path)
            logger.trace(
                f"Getting identifier from metadata_name:{self.name!r} path:{path!r} item:{item!r} listingitem:{listingitem!r}"
            )
            identifier = self.get_identifier(item)
            if not identifier:
                logger.trace(
                    f"We found no identifier for path:{path!r} item:{item!r} listingitem:{listingitem!r}"
                )
                missing_metadata.append((path, item, listingitem))
                continue

            metadata_mapping[path].add(identifier)
            required_metadata.add(identifier)
            if has_name:
                identifier_name_mapping[identifier] = item.id

        resolution_links_created = False
        search_resolve = self.config.get("search_resolve")
        if missing_metadata and search_resolve and self.metadata_resolution_link_model:
            names = {
                item.id: (path, item, listingitem)
                for path, item, listingitem in missing_metadata
            }
            resolution_links = self.metadata_resolution_link_model.objects.select_related(
                "metadata"
            ).filter(
                search_resolve=search_resolve, name__in=names.keys()
            )

            logger.debug(f"We have {len(resolution_links)} resolution links")
            missing_resolution_links = {}
            for resolution_link in resolution_links:
                if resolution_link.name not in names:
                    logger.warning(
                        f"Name {resolution_link.name} not found in resolution link query"
                    )
                    continue

                path, item, listingitem = names.pop(resolution_link.name)
                if resolution_link.metadata:
                    identifier = resolution_link.metadata.identifier
                    metadata_mapping[path].add(identifier)
                    logger.trace(
                        f"Solving path:{path} to identifier:{identifier} using resolution links"
                    )
                    required_metadata.add(identifier)
                    if has_name:
                        identifier_name_mapping[identifier] = item.id
                elif listingitem:
                    missing_resolution_links[listingitem.path] = resolution_link

            if names or missing_resolution_links:
                logger.info("Missing resolution links, creating")
                resolution_links_create = []
                for name in names.keys():
                    resolution_links_create.append(
                        self.metadata_resolution_link_model(
                            search_resolve=search_resolve,
                            name=name,
                            plugin=self._plugin_obj,
                        )
                    )

                self.metadata_resolution_link_model.objects.bulk_create(
                    resolution_links_create
                )

                resolution_links = self.metadata_resolution_link_model.objects.filter(
                    search_resolve=search_resolve, name__in=names.keys()
                )

                mapping = {}

                for resolution_link in resolution_links:
                    path, item, listingitem = names.pop(resolution_link.name)
                    mapping[path] = resolution_link

                with transaction.atomic():
                    for listingitem in listing_item_model.objects.filter(
                        app=listing_item_root.app,
                        path__in=list(mapping.keys())
                        + list(missing_resolution_links.keys()),
                    ):
                        resolution_link = (
                            listingitem.path in mapping
                            and mapping.pop(listingitem.path)
                            or missing_resolution_links.pop(listingitem.path)
                        )
                        resolution_link.listingitems.add(listingitem)

                resolution_links_created = True

        metadata = self.model.objects.filter(identifier__in=required_metadata)

        if fetch_metadata or self.prepopulate_metadata:
            initial_status = "pending"
            metadata.filter(last_update_status="do-not-fetch").update(
                last_update_status=initial_status
            )
        else:
            initial_status = "do-not-fetch"

        metadata = dict(metadata.values_list("identifier", "pk"))

        missing_metadata = required_metadata - set(metadata.keys())

        if missing_metadata or resolution_links_created:
            if missing_metadata:
                if has_name:
                    self.model.objects.bulk_create(
                        self.model(
                            identifier=identifier,
                            last_update_status=initial_status,
                            plugin=self._plugin_obj,
                            name=identifier_name_mapping[identifier],
                        )
                        for identifier in missing_metadata
                    )
                else:
                    self.model.objects.bulk_create(
                        self.model(
                            identifier=identifier,
                            plugin=self._plugin_obj,
                            last_update_status=initial_status,
                        )
                        for identifier in missing_metadata
                    )

            if self.prepopulate_metadata:
                logger.debug(
                    "This metadata type should be updated before we use it, waiting for update to complete"
                )
                for _ in range(20 * 10):
                    if self.update():
                        break
                    else:
                        time.sleep(0.1)
            else:
                self.schedule_update()

        metadata.update(
            self.model.objects.filter(identifier__in=missing_metadata).values_list(
                "identifier", "pk"
            )
        )

        listingitem_metadata_mapping = defaultdict(set)
        for path, pk in listing_item_model.objects.filter(
            Q(pk=listing_item_root.pk) | Q(parent=listing_item_root)
        ).values_list("path", "pk"):
            for identifier in metadata_mapping[path]:
                listingitem_metadata_mapping[pk].add((metadata[identifier], None))

        self.listingitem_relinking(listing_item_root, listingitem_metadata_mapping)

        if index_writer and self.fulltext_fields:
            logger.debug(f"Creating search index for metadata {self.name}")

            obj_type = f"metadata_{self.plugin_name}__metadata__"
            last_path, used_searchstr = None, set()
            query = listing_item_root.listingitem_set.filter(
                Q(
                    path__in=set(item_creation_mapping.keys())
                    - set(existing_listing_items.keys())
                )
                | Q(
                    path__in=existing_listing_items.keys(),
                    **{obj_type + "last_updated": listing_last_update},
                )
            ).distinct()

            for searchstrs in query.values_list(
                "path", *[(obj_type + x) for x in self.fulltext_fields]
            ).order_by("pk"):
                searchstrs = list(searchstrs)
                path = searchstrs.pop(0)
                if path != last_path:
                    used_searchstr = set()
                    last_path = path

                for searchstr in searchstrs:
                    if searchstr in used_searchstr or not searchstr:
                        continue

                    used_searchstr.add(searchstr)
                    index_writer.index(path, searchstr)

    def ready(self):
        self.schedule_update()

    def schedule_update(self):
        if not self._job_schedule:
            self._job_schedule = settings.SCHEDULER.add_job(
                self.update_request, trigger="interval", minutes=30
            )

        self.update_request()

    def update_request(self, *args):
        threadify(self.update)()

    def update(self):
        """Check for metadata to update"""
        if self.currently_updating_lock.locked():
            logger.info("Updater lock already locked, bailing")
            return False

        logger.debug(f"Trying to update metadata for {self.name}")
        with self.currently_updating_lock:
            if self.metadata_resolution_link_model:
                while not self._do_unload and not reactor._stopped:
                    resolution_links = self.metadata_resolution_link_model.objects.filter(
                        last_update_status="pending", listingitems__isnull=False
                    )
                    if not resolution_links:
                        break
                    resolution_link = resolution_links[0]

                    logger.debug(f"Resolving link {resolution_link!r}")

                    resolution_link.last_update_status = "updating"
                    resolution_link.save()

                    if resolution_link.plugin:
                        p = resolution_link.plugin.get_plugin()
                        config = p.config
                    else:
                        config = self.config

                    resolution_link.populated = True
                    try:
                        identifier = resolution_link.resolve(config)
                    except:
                        logger.exception(f"Failed to resolve link {resolution_link!r}")
                        resolution_link.last_update_status = "failed"
                    else:
                        logger.debug(
                            f"Successfully resolved link {resolution_link!r} to {identifier}"
                        )
                        resolution_link.last_update_status = "success"

                        if identifier:
                            metadata, _ = self.model.objects.get_or_create(
                                identifier=identifier,
                                defaults={"plugin": self._plugin_obj},
                            )
                            logger.debug(
                                f"Got identifier {identifier}, linking all applicable listingitems"
                            )
                            resolution_link.metadata = metadata

                            for listingitem in resolution_link.listingitems.all():
                                self.listing_item_relation_model.objects.create(
                                    listingitem=listingitem, metadata=metadata
                                )

                    resolution_link.last_updated = now()
                    resolution_link.save()

            while not self._do_unload and not reactor._stopped:
                metadatas = self.model.objects.filter(
                    last_update_status="pending"
                ).order_by("-priority")
                if not metadatas:
                    break
                metadata = metadatas[0]

                logger.debug(f"Populating {metadata!r}")

                metadata.last_update_status = "updating"
                metadata.save()

                if metadata.plugin:
                    p = metadata.plugin.get_plugin()
                    config = p.config
                else:
                    config = self.config

                metadata.populated = True
                try:
                    metadata.populate(config)
                    metadata.prefetch_related_denormalized = self.get_prefetch_related_denormalized(
                        metadata
                    )
                except:
                    logger.exception(f"Failed to populate metadata {metadata!r}")
                    metadata.last_update_status = "failed"
                else:
                    logger.debug(f"Successfully populated metadata {metadata!r}")
                    metadata.last_update_status = "success"

                metadata.last_updated = now()
                metadata.save()

                if metadata.identifier in self.event_queue:
                    self.event_queue[metadata.identifier].set()

        return True

    def unload(self):
        self._do_unload = True

        if self._job_schedule:
            self._job_schedule.remove()


class BaseMetadataSerializer(serializers.ModelSerializer):
    id = serializers.CharField(source="identifier", read_only=True)

    class Meta:
        model = None
        exclude = (
            "last_updated",
            "created",
            "identifier",
            "priority",
            "plugin",
            "prefetch_related_denormalized",
        )


class MetadataSerializer(CoverSerializerMixin, BaseMetadataSerializer):
    cover = serializers.SerializerMethodField("get_cover_url")



import logging

from malparser import MAL  # NOQA
from rest_framework import serializers

from ...bases.metadata.remotemetadata import (
    MetadataSerializer,
    RemoteMetadataHandlerPlugin,
)
from .filters import MetadataFilter
from .models import ListingItemRelation, Metadata

logger = logging.getLogger(__name__)


class MALSeasonSerializer(serializers.Serializer):
    year = serializers.IntegerField()
    season = serializers.CharField()


class MALMetadataSerializer(MetadataSerializer):
    anime_type = serializers.StringRelatedField()
    tv_rating = serializers.StringRelatedField()
    genres = serializers.StringRelatedField(many=True)
    season = MALSeasonSerializer()
    year = serializers.SerializerMethodField()

    class Meta(MetadataSerializer.Meta):
        model = Metadata

    def get_year(self, obj):
        if obj.season and obj.season.year:
            return obj.season.year


class MALMetadataHandlerPlugin(RemoteMetadataHandlerPlugin):
    plugin_name = "mal"
    select_related = ["season", "tv_rating", "anime_type"]
    prefetch_related = ["genres", "alternativetitle_set", "related"]
    serializer = MALMetadataSerializer
    model = Metadata
    filter = MetadataFilter
    listing_item_relation_model = ListingItemRelation
    fulltext_fields = ["title"]

    __traits__ = ["primary_metadata", "metadata_anime"]



from django.conf.urls import url
from unplugged import RelatedPluginField, Schema, ServicePlugin

from ...plugins import ConfigPlugin
from .views import ConfigView


class ConfigServiceSchema(Schema):  # Default configs
    config = RelatedPluginField(plugin_type=ConfigPlugin, required=True)


class ConfigServicePlugin(ServicePlugin):
    plugin_name = "config"
    config_schema = ConfigServiceSchema

    def __init__(self, config):
        self.config = config["config"]

    def get_urls(self):
        return [url("^/?$", ConfigView.as_view(service=self))]



from abc import abstractmethod, abstractproperty

from unplugged import PluginBase


class MetadataHandlerPlugin(PluginBase):
    plugin_type = "metadatahandler"

    @abstractproperty
    def linkable(self):
        """
        Should the metadata be linked with other metadata handlers.
        """
        raise NotImplementedError

    @abstractmethod
    def get_metadata(self, request, identifier):
        """
        Returns an instance of matching metadata
        """
        raise NotImplementedError

    @abstractmethod
    def link_metadata_listingitems(self, listing_build, fetch_metadata=True):
        """
        Links metadata and listingitems.
        """
        raise NotImplementedError

    @abstractmethod
    def populate_metadata_jsonapi(self, request, root):
        """
        Adds metadata to a JSONAPI response.
        """
        raise NotImplementedError



# Generated by Django 2.2.8 on 2019-12-13 16:29

import jsonfield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ListingCache",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app", models.CharField(max_length=30)),
                ("searcher_name", models.CharField(max_length=50)),
                ("search_token", models.CharField(max_length=50)),
                ("path", models.CharField(default="", max_length=500)),
                ("listing", jsonfield.fields.JSONField()),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
            options={
                "unique_together": {("app", "searcher_name", "search_token", "path")},
            },
        ),
    ]



from urllib.parse import quote

from django.db import models
from jsonfield import JSONField
from thomas import Item, router


class ListingItem(models.Model):
    """An item in a listing view"""

    app = models.CharField(max_length=30)
    name = models.CharField(max_length=200, db_index=True, default="")
    path = models.CharField(max_length=500)  # absolute path
    parent = models.ForeignKey(
        "self", null=True, db_index=True, on_delete=models.deletion.CASCADE
    )
    datetime = models.DateTimeField()
    config = JSONField(default=dict)
    attributes = JSONField(default=dict)

    LISTINGITEM_TYPE_FILE = "file"
    LISTINGITEM_TYPE_FOLDER = "folder"
    LISTINGITEM_TYPES = (
        (LISTINGITEM_TYPE_FILE, "File"),
        (LISTINGITEM_TYPE_FOLDER, "Folder"),
    )
    item_type = models.CharField(max_length=10, choices=LISTINGITEM_TYPES)
    is_root = models.BooleanField(default=False)
    last_updated = models.DateTimeField(null=True)
    last_checked = models.DateTimeField(null=True)

    class Meta:
        ordering = ("datetime",)
        unique_together = (("app", "path"),)

    def __str__(self):
        return self.path

    def get_absolute_url(self):
        url = "/%s/%s" % (self.app, self.path)
        return quote(url.encode("utf-8"))

    def get_name(self):
        return self.path.split("/")[-1]

    def get_full_path(self):
        if self.path:
            return "%s/%s" % (self.app, self.path)
        else:
            return self.app

    def get_local_path(self):
        return "/".join(self.path.split("/")[1:])

    def get_metadata(self, metadata):
        related = getattr(self, "metadata_%s" % (metadata,), None)
        if not related:
            return []

        return [m.metadata for m in related.all().select_related("metadata")]

    def get_all_related_metadata(self):
        metadatas = []

        for ro in self._meta.related_objects:
            if not ro.name.startswith("metadata_"):
                continue

            related = getattr(self, ro.name, None)
            metadatas.extend(
                [m.metadata for m in related.all().select_related("metadata")]
            )

        return metadatas

    def get_original_item(self):
        return Item.unserialize(self.config["original_item"], router=router)



from timeoutthreadpoolexecutor import TimeoutThreadPoolExecutor  # isort:skip
from concurrent.futures import thread  # isort:skip

thread.ThreadPoolExecutor = TimeoutThreadPoolExecutor  # isort:skip

import asyncio
import logging
import logging.handlers
import os
import sys
import time

import daphne.server
import django
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from channels.routing import get_default_application
from txasgiresource import ASGIResource
from zope.interface import implementer

from twisted.application import service
from twisted.application.service import IServiceMaker
from twisted.internet import endpoints, reactor
from twisted.plugin import IPlugin
from twisted.python import log, usage
from twisted.web import resource, server, static

os.environ["DJANGO_SETTINGS_MODULE"] = "main.settings"

logger = logging.getLogger(__name__)


class Options(usage.Options):
    optFlags = [
        ["trace", "t", "Enable trace logging"],
        ["debug", "d", "Enable debug logging"],
        ["stdout", "s", "Log to stdout"],
        ["djangodebug", "j", "Enable django debug"],
        ["clearcache", "c", "Clear caching on every action"],
    ]

    optParameters = [
        [
            "description",
            "e",
            "tcp:45477:interface=0.0.0.0",
            "Twisted server description",
        ],
        [
            "reverseproxy",
            "r",
            "smart",
            "Set reverse proxy mode (uses x-forwarded headers), values are on, off and smart",
        ],
    ]


class Root(resource.Resource):
    def __init__(self, asgi_resource):
        resource.Resource.__init__(self)
        self.asgi_resource = asgi_resource

    def getChild(self, path, request):
        path0 = request.prepath.pop(0)
        request.postpath.insert(0, path0)
        return self.asgi_resource


class ASGIService(service.Service):
    def __init__(self, site, resource, description):
        self.site = site
        self.resource = resource
        self.description = description

    def startService(self):
        self.endpoint = endpoints.serverFromString(reactor, self.description)
        self.endpoint.listen(self.site)

    def stopService(self):
        self.resource.stop()


class SchedulerService(service.Service):
    def __init__(self):
        self.scheduler = AsyncIOScheduler()

    def startService(self):
        self.scheduler.start()

    def stopService(self):
        self.scheduler.shutdown()


class File(static.File):
    def directoryListing(self):
        return self.forbidden


@implementer(IServiceMaker, IPlugin)
class ServiceMaker(object):
    tapname = "tridentstream"
    description = "Tridentstream media server"
    options = Options

    def makeService(self, options):
        TRACE_LEVEL_NUM = 5
        logging.addLevelName(TRACE_LEVEL_NUM, "TRACE")

        def trace(self, message, *args, **kws):
            self._log(TRACE_LEVEL_NUM, message, args, **kws)

        logging.Logger.trace = trace

        import urllib3  # TODO: remove this code again, not sure why it doesn't just fix it by itself.

        urllib3.disable_warnings()

        if options["djangodebug"]:
            os.environ["DJANGO_DEBUG"] = "1"

        logging.getLogger("rebulk.rules").setLevel(logging.WARNING)
        logging.getLogger("rebulk.rebulk").setLevel(logging.WARNING)
        logging.getLogger("chardet.charsetprober").setLevel(logging.WARNING)

        if options["trace"]:
            log_level = TRACE_LEVEL_NUM
            asyncio.set_event_loop(reactor._asyncioEventloop)
            asyncio.get_event_loop().set_debug(True)
        elif options["debug"]:
            log_level = logging.DEBUG
        else:
            log_level = logging.WARNING

        if options["stdout"]:
            handler = logging.StreamHandler(sys.stdout)
        else:
            handler = logging.handlers.RotatingFileHandler(
                "tridentstream.log", maxBytes=50 * 1024 * 1024, backupCount=5
            )

        handler.setLevel(log_level)
        formatter = logging.Formatter(
            "%(asctime)-15s:%(levelname)s:%(name)s:%(lineno)d:%(message)s"
        )
        handler.setFormatter(formatter)
        root = logging.getLogger()
        root.setLevel(log_level)
        root.addHandler(handler)

        observer = log.PythonLoggingObserver()
        observer.start()

        from django.conf import settings

        if options["clearcache"]:
            settings.DEBUG_CLEAR_CACHE = True
            from django.core.cache import cache

            cache.clear()

        multi = service.MultiService()

        django.setup()
        application = get_default_application()

        if options["reverseproxy"] == "on":
            asgiresource = ASGIResource(
                application, use_proxy_headers=True, use_x_sendfile=True
            )
        elif options["reverseproxy"] == "off":
            asgiresource = ASGIResource(
                application, use_proxy_proto_header=True, use_x_sendfile=True
            )
        else:
            asgiresource = ASGIResource(
                application, automatic_proxy_header_handling=True, use_x_sendfile=True
            )
        r = Root(asgiresource)

        from thomas import OutputBase

        http_output_cls = OutputBase.find_plugin("http")
        http_output = http_output_cls(url_prefix="stream")
        http_output.start()
        r.putChild(b"stream", http_output.resource)

        settings.THOMAS_HTTP_OUTPUT = http_output

        from django.contrib import admin

        r.putChild(
            settings.STATIC_URL.strip("/").encode("utf-8"),
            File(
                os.path.join(os.path.dirname(django.contrib.admin.__file__), "static")
            ),
        )

        # Gluing it all together
        site = server.Site(r)

        multi.addService(ASGIService(site, asgiresource, options["description"]))

        scheduler_service = SchedulerService()
        multi.addService(scheduler_service)
        settings.SCHEDULER = scheduler_service.scheduler

        def initialize():
            def _initialize():
                from unplugged.bootstrap import bootstrap_all

                bootstrap_all()

            loop = asyncio.get_running_loop()
            asyncio.ensure_future(loop.run_in_executor(None, _initialize))

        reactor.callLater(0, initialize)

        # Django doesn't seem to want to kill connections
        def cleanup_database_connections():
            def cleanup_thread():
                from django.db import close_old_connections

                close_old_connections()

            settings.SCHEDULER.add_job(cleanup_thread, "interval", minutes=15)

        reactor.callLater(0, cleanup_database_connections)

        return multi


tridentstream = ServiceMaker()



import logging
import uuid

from rest_framework import serializers

from ...bases.metadata.mixins import CoverSerializerMixin, PopulateMetadataJSONAPIMixin
from ...bases.metadata.schema import MetadataSchema
from ...plugins import MetadataHandlerPlugin
from .models import Availability, ListingItemAvailability, ListingItemRelation

logger = logging.getLogger(__name__)


class AvailableSerializer(serializers.ModelSerializer):
    id = serializers.SerializerMethodField("get_metadata_identifier")
    type = serializers.SerializerMethodField("get_metadata_type")

    class Meta:
        model = Availability
        exclude = ("app", "identifier")

    def get_metadata_identifier(self, obj):
        return obj.pk

    def get_metadata_type(self, obj):
        return f"metadata_{obj.metadata_name}"

    def get_populated(self, obj):
        return True


class AvailableMetadataHandlerPlugin(
    PopulateMetadataJSONAPIMixin, MetadataHandlerPlugin
):
    plugin_name = "available"
    __traits__ = ["availability"]
    priority = -10
    config_schema = MetadataSchema
    linkable = False

    model = Availability
    filter = None
    listing_item_relation_model = ListingItemRelation
    select_related = []
    prefetch_related = []
    serializer = AvailableSerializer
    metadata_embed_method = "relate"  # choices: include, relate, embed

    def get_metadata(self, request, identifier):
        """
        Not something this type of metadata can actually handle
        """
        pass

    def link_metadata_listingitems(self, listing_build, fetch_metadata=True):
        item_creation_mapping = listing_build.item_creation_mapping
        listingitem_mapping = listing_build.listingitem_mapping

        ListingItemRelation.objects.filter(
            listingitem__in=listingitem_mapping.values()
        ).delete()  # TODO: optimize
        objs = []
        availability_mapping = {}
        for path, item in item_creation_mapping.items():
            if path not in listingitem_mapping:
                continue

            listingitem = listingitem_mapping[path]
            for availability in item.get("metadata:availability", []):
                objs.append(
                    ListingItemAvailability(
                        app=availability["app"],
                        identifier=availability["identifier"],
                        listingitem=listingitem,
                    )
                )

                availability_mapping.setdefault(availability["app"], {}).setdefault(
                    str(availability["identifier"]), []
                ).append(listingitem)

        ListingItemAvailability.objects.bulk_create(objs)

        objs = []
        for app, identifiers in availability_mapping.items():
            for availability in self.model.objects.filter(
                app=app, identifier__in=identifiers.keys()
            ):
                for listingitem in identifiers[availability.identifier]:
                    objs.append(
                        self.listing_item_relation_model(
                            metadata=availability, listingitem=listingitem
                        )
                    )

        self.listing_item_relation_model.objects.bulk_create(objs)

    # Availability trait
    def mark_available(self, app, identifier):
        """
        Mark this app/identifier as available.
        """
        availability, _ = Availability.objects.get_or_create(
            app=app, identifier=identifier
        )

        objs = []
        for lia in ListingItemAvailability.objects.filter(
            app=app, identifier=identifier
        ):
            objs.append(
                self.listing_item_relation_model(
                    listingitem=lia.listingitem, metadata=availability
                )
            )

        self.listing_item_relation_model.objects.bulk_create(objs)

    def populate_available(self, item):
        """
        Loop through the item and mark items we know are available as available.
        """
        availability_to_item_mapping = {}

        def _expand_item(item):
            for availability in item.get("metadata:availability", []):
                availability_to_item_mapping.setdefault(
                    availability["app"], {}
                ).setdefault(str(availability["identifier"]), []).append(item)

            if item.is_expanded:
                for sub_item in item.list():
                    _expand_item(sub_item)

        _expand_item(item)

        for app, identifiers in availability_to_item_mapping.items():
            for availability in Availability.objects.filter(
                app=app, identifier__in=identifiers.keys()
            ):
                for item in identifiers[availability.identifier]:
                    self.set_available(item)

    def get_recently_made_available(self, app, since):
        """
        Return a list of app, item, datetime dict that contains
        all made available by app since since.
        """
        return [
            {
                "app": available.app,
                "identifier": available.identifier,
                "datetime": available.created,
            }
            for available in Availability.objects.filter(app=app, created__gte=since)
        ]

    def set_available(self, item):
        """Sets the correct metadata key to indicate availability"""
        logger.trace(f"Setting item:{item!r} as available")
        item.setdefault("metadata", {})["available"] = True



import logging

from django.conf.urls import url
from unplugged import Schema, ServicePlugin

from .views import ImageCacheView

logger = logging.getLogger(__name__)


class ImageCacheServicePlugin(ServicePlugin):
    plugin_name = "imgcache"
    config_schema = Schema

    __traits__ = ["image_cache"]

    def get_urls(self):
        return [
            url("^(?P<imgcache_name>[-\w]+)/?$", ImageCacheView.as_view(service=self))
        ]



import logging

from django.db import models

from ...bases.metadata.models import BaseListingItemRelation, BaseMetadataLink

logger = logging.getLogger(__name__)


class FirstSeen(BaseMetadataLink):
    """Metadata related to what the user saw"""

    datetime = models.DateTimeField(auto_now_add=True)

    @property
    def metadata_name(self):
        return "firstseen"

    @property
    def identifier(self):
        return self.metadata.identifier


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(FirstSeen, on_delete=models.deletion.CASCADE)



import logging

from django.apps import AppConfig as DjangoAppConfig

logger = logging.getLogger(__name__)


class AppConfig(DjangoAppConfig):
    name = "tridentstream.dbs.lmdb"
    verbose_name = "LMDB Database Plugin"
    label = "db_lmdb"

    def ready(self):
        try:
            from .handler import LMDBDatabasePlugin  # NOQA
        except ImportError:
            logger.info("Unable to load the LMDB database plugin")



import logging
from urllib.parse import urljoin

import requests
from thomas import Item, StreamerBase, router
from unplugged import Schema, fields

from twisted.internet import threads

from ...exceptions import NotModifiedException, PathNotFoundException
from ...plugins import InputPlugin
from ...stream import Stream

logger = logging.getLogger(__name__)


class RemoteFilesystemInputSchema(Schema):
    url = fields.String()
    token = fields.String()

    priority = fields.Integer(default=5)


class RemoteFilesystemStreamer:
    def __init__(self, plugin, path):
        self.plugin = plugin
        self.path = path

    def evaluate(self):
        return self.plugin.config["priority"] + 1

    def stream(self):
        return self.plugin.stream(self.path)


class RemoteFilesystemInputPlugin(InputPlugin):
    plugin_name = "remotefilesystem"

    config_schema = RemoteFilesystemInputSchema

    simpleadmin_templates = True

    def __init__(self, config):
        self.config = config

        self.route_input_rfs_list = f"input_rfs_list_{self.name}"
        router.register_handler(
            self.route_input_rfs_list, self.thomas_list, False, True, False
        )

        self.route_input_rfs_stream = f"input_rfs_stream_{self.name}"
        router.register_handler(
            self.route_input_rfs_stream, self.thomas_stream, False, False, True
        )

    def unload(self):
        router.unregister_handler(self.route_input_rfs_list)
        router.unregister_handler(self.route_input_rfs_stream)

    def get_headers(self):
        return {"Authorization": f"Token {self.config['token']}"}

    def get_item(self, path):
        item = Item(id=path.strip().split("/")[-1], router=router)

        item.expandable = True
        item.streamable = True
        self.add_routes(item, path, skip=True)
        # item.add_route(self.route_input_rfs_list, False, True, False, kwargs={'path': path})

        # item.streamable = True
        # item.add_route(self.route_input_rfs_stream, False, False, True, kwargs={'path': path})

        return item

    def add_routes(self, item, path, skip=False):
        if not skip:
            if path:
                path = f"{path}/{item.id}"
            else:
                path = item.id

        if item.is_streamable:
            item.add_route(
                self.route_input_rfs_stream, False, False, True, kwargs={"path": path}
            )

        if item.is_listable:
            if item.is_expanded:
                for nested_item in item.nested_items:
                    self.add_routes(nested_item, path)
            else:
                item.add_route(
                    self.route_input_rfs_list, False, True, False, kwargs={"path": path}
                )

    def thomas_list(self, item, path, depth=0, modified_since=None):
        logger.info(f"Listing path {path!r} with depth {depth}")
        item_id = item.id
        headers = self.get_headers()
        if modified_since:
            headers["If-Modified-Since"] = modified_since.strftime(
                "%a, %d %b %Y %H:%M:%S GMT"
            )

        r = requests.get(
            urljoin(self.config["url"].strip("/") + "/", path),
            params={"depth": depth},
            headers=headers,
        )

        if r.status_code == 200:
            item = Item.unserialize(r.json(), router=router)
            item.id = item_id
            self.add_routes(item, path, skip=True)
            return item
        elif r.status_code == 304:
            raise NotModifiedException()
        elif r.status_code == 404 or r.status_code == 403:
            raise PathNotFoundException()
        else:
            logger.warning(
                f"Unknown status code {r.status_code} while listing {self.name}/{path}"
            )

    def thomas_stream(self, item, path):
        logger.info(f"Trying to stream {path!r}")
        return RemoteFilesystemStreamer(self, path)

    def stream(self, path):
        logger.info(f"Trying to stream {path!r}")
        headers = self.get_headers()
        r = requests.post(
            urljoin(self.config["url"].strip("/") + "/", path), headers=headers
        )

        if r.status_code != 200:
            raise PathNotFoundException()

        return Stream.unserialize(r.json())



from abc import abstractmethod

from unplugged import PluginBase


class HistoryPlugin(PluginBase):
    plugin_type = "history"

    @abstractmethod
    def log_history(self, config, listingitem, viewstate):
        """
        Log a user who watched a listingitem with viewstate, update viewstate if got info.
        """
        raise NotImplementedError



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.inputs.fs"
    verbose_name = "Filesystem Input"
    label = "inputs_fs"

    def ready(self):
        from .handler import FilesystemInputPlugin  # NOQA



# import os
# import shutil

# from django.contrib.auth.models import User
# from django.test import TestCase

# from freezegun import freeze_time

# from ....dbs.memory.handler import MemoryDatabasePlugin
# from ..handler import FilesystemInputPlugin


# def test_filesystem_basic(tmpdir):
#     d = tmpdir.mkdir('data')
#     d.mkdir('subdata')

#     db = MemoryDatabasePlugin({})
#     fs = FilesystemInputPlugin({
#         'db': db,
#         'paths': [
#             {'path': str(d), 'virtual_root': ''}
#         ]
#     })

#     with freeze_time("2018-01-01"):
#         yield fs.rescan()

#     raise Exception(fs.list('', 10).serialize())


# class FilesystemInputHandlerTestCase(TestCase):
#     maxDiff = 5000
#     def setUp(self):
#         self.reactor = threaded_reactor()
#         self.temppath = tempfile.mkdtemp()

#         os.mkdir(os.path.join(self.temppath, 'data1'))
#         os.mkdir(os.path.join(self.temppath, 'data2'))
#         os.mkdir(os.path.join(self.temppath, 'data3'))
#         self.db_1 = MemoryDatabasePlugin({})
#         config = {
#             'db': self.db_1,
#             'paths': [
#                 {'path': os.path.join(self.temppath, 'data1'), 'virtual_root': 'data1'},
#                 {'path': os.path.join(self.temppath, 'data2'), 'virtual_root': 'data2'},
#             ]
#         }
#         self.fs_1 = FilesystemInputPlugin(config)

#         os.mkdir(os.path.join(self.temppath, 'data1', 'folder1'))
#         with open(os.path.join(self.temppath, 'data1', 'folder1', 'file1'), 'wb') as f:
#             f.write(b' ' * 20)

#         os.mkdir(os.path.join(self.temppath, 'data3', 'folder3'))
#         with open(os.path.join(self.temppath, 'data3', 'folder3', 'file3'), 'wb') as f:
#             f.write(b' ' * 30)

#         os.symlink(os.path.join(self.temppath, 'data3', 'folder3'), os.path.join(self.temppath, 'data1', 'folder3'))

#         os.mkdir(os.path.join(self.temppath, 'data1', 'folder2'))

#     def tearDown(self):
#         if self.temppath.startswith('/tmp'):
#             shutil.rmtree(self.temppath)

# @deferred()
# @defer.inlineCallbacks
# def test_rescan(self):
#     yield self.fs_1._rescan()

#     os.remove(os.path.join(self.temppath, 'data1', 'folder1', 'file1'))
#     os.rmdir(os.path.join(self.temppath, 'data1', 'folder1'))

#     yield self.fs_1._rescan()

#     listing = yield self.fs_1.list('', 3)
#     from pprint import pprint; pprint(listing.serialize())

#     raise Exception()



from unplugged import Schema

from ...plugins import ConfigPlugin
from .models import DefaultSetting, Setting


class ConfigConfigPlugin(ConfigPlugin):
    plugin_name = "config"
    config_schema = Schema

    def __init__(self, config):
        super(ConfigConfigPlugin, self).__init__(config)

        self.schemas = {}

    def get_default_config(self, namespace, key):
        for default_type in self.default_types:
            try:
                return DefaultSetting.objects.get(
                    default_type=default_type, namespace=namespace, key=key
                ).value
            except DefaultSetting.DoesNotExist:
                pass
        return None

    def get_user_config(self, user, namespace, key):
        try:
            return Setting.objects.get(user=user, namespace=namespace, key=key).value
        except Setting.DoesNotExist:
            return self.get_default_config(namespace, key)

    def set_default_config(self, default_type, namespace, key, value):
        try:
            s = DefaultSetting.objects.get(
                default_type=default_type, namespace=namespace, key=key
            )
        except DefaultSetting.DoesNotExist:
            s = DefaultSetting(default_type=default_type, namespace=namespace, key=key)

        if value is None:
            s.delete()
        else:
            s.value = value
            s.save()

    def set_user_config(self, user, namespace, key, value):
        try:
            s = Setting.objects.get(user=user, namespace=namespace, key=key)
        except Setting.DoesNotExist:
            s = Setting(user=user, namespace=namespace, key=key)

        if value is None:
            s.delete()
        else:
            s.value = value
            s.save()

    def set_config_schema(self, namespace, key, schema):
        self.schemas[(namespace, key)] = schema

    def get_config_schema(self, namespace, key):
        return self.schemas.get((namespace, key))



import rest_framework_filters as filters

from .models import ListingItemRelation


class MetadataFilter(filters.FilterSet):
    id = filters.AllLookupsFilter(
        field_name="metadata__identifier", name="metadata__identifier"
    )
    year = filters.AllLookupsFilter(field_name="metadata__year", name="metadata__year")
    season = filters.AllLookupsFilter(
        field_name="metadata__season", name="metadata__season"
    )
    episode = filters.AllLookupsFilter(
        field_name="metadata__episode", name="metadata__episode"
    )
    fix_type = filters.AllLookupsFilter(
        field_name="metadata__fix_type", name="metadata__fix_type"
    )

    class Meta:
        model = ListingItemRelation
        order_by = ["year", "season", "episode", "fix_type"]
        fields = []



import logging

from django.contrib.auth.models import User
from unplugged import RelatedPluginField, Schema, fields
from wampyre.realm import realm_manager

from ...plugins import NotifierPlugin

logger = logging.getLogger(__name__)


class MultiNotifierSchema(Schema):
    notifiers = fields.List(
        RelatedPluginField(plugin_type=NotifierPlugin), many=True, default=list
    )


class MultiNotifierNotifierHandlerPlugin(NotifierPlugin):
    plugin_name = "multinotifier"
    config_schema = MultiNotifierSchema

    def notify(self, notification):
        for notifier in self.config.get("notifiers", []):
            notifier.notify(notification)



import logging
from datetime import datetime, timedelta

import pytz
from django.conf import settings
from django.db import transaction
from django.utils.timezone import now
from thomas import Item, router

from ...exceptions import NotModifiedException, PathNotFoundException
from ...locktracker import LockTracker
from .models import ListingItem

logger = logging.getLogger(__name__)


class ListingBuild:
    config = None
    index_writer = None
    listing_item_root = None
    existing_listing_items = None
    item_creation_mapping = None
    listingitem_mapping = None
    linkable_metadata = None
    listing_last_update = None


class ListingBuilder:
    def __init__(self, service):
        self.service = service
        self.locktracker = LockTracker()

    def get_item(self, path):
        """
        Returns item at a given path if it exists.
        Cannot try to build
        """
        try:
            return ListingItem.objects.get(app=self.service.name, path=path)
        except ListingItem.DoesNotExist:
            return None

    def get_listing(
        self, config, path, use_background_recheck=True, do_not_rebuild=False
    ):
        """
        Returns listing if it is good and available
        """
        path = path.strip("/")

        last_modified = None
        try:
            listing_item_root = ListingItem.objects.get(
                app=self.service.name, path=path
            )
            if listing_item_root.is_root:
                last_modified = listing_item_root.last_updated

                if settings.DEBUG_CLEAR_CACHE:
                    logger.debug("Clear cache enabled, removing old listing")
                    last_modified = None
                    ListingItem.objects.filter(parent=listing_item_root).delete()
                    listing_item_root.delete()
                    listing_item_root = None
        except ListingItem.DoesNotExist:
            listing_item_root = None

        if (
            config["level"].get("background_recheck")
            and use_background_recheck
            and listing_item_root
            and listing_item_root.last_checked
            and listing_item_root.last_checked
            > now() - timedelta(minutes=self.service.automatic_rebuild_timer)
        ):
            logger.info(
                "We run updates to this listing in the background and it is sufficiently recent to be used"
            )
            if not do_not_rebuild:
                self.service.rebuild_listing(config, path, delay=True)
            return listing_item_root

        if listing_item_root:
            item = Item.unserialize(listing_item_root.config["original_item"])
        else:
            item = self.service.get_item(config, path)

        try:
            if not item:
                raise PathNotFoundException()

            item.list(depth=config["level"]["listing_depth"])
            if last_modified and item.modified and item.modified <= last_modified:
                raise NotModifiedException()
        except NotModifiedException:
            logger.info("Our current listing is up-to-date, no need to do anything")
            listing_item_root.last_checked = now()
            listing_item_root.save()
            return listing_item_root
        except PathNotFoundException:
            logger.info(
                "We got an all paths not found, just returning empty for now and hope it will be better next time"
            )
            return None

        if not item.is_listable:
            if last_modified is None:
                logger.info(
                    "Empty listing and no last_modified, this is an unknown path"
                )
                return None
            else:
                logger.info(
                    "Empty listing and last_modified, returning already built root"
                )
                return listing_item_root

        lock_path = "%s/%s" % (self.service.name, path)
        with self.locktracker.get_path(lock_path) as lt:
            if lt.waited:
                logger.info(
                    "Seems like someome built this view just before us, no need to build"
                )
                try:
                    listing_item_root = ListingItem.objects.get(
                        app=self.service.name, path=path
                    )
                except ListingItem.DoesNotExist:
                    listing_item_root = None
            else:
                logger.info("Got lock, building listview")
                listing_item_root = self.build_listing(item, config, path)

        return listing_item_root

    def build_listing(self, listing, config, path):
        """
        Builds a listing view using given specifications.

        :param path: Root of listing view.
        """
        listing_build = ListingBuild()
        listing_build.config = config

        try:
            listing_item_root = ListingItem.objects.get(
                app=self.service.name, path=path
            )
        except:
            listing_item_root = ListingItem(app=self.service.name, path=path)

        listing_build.listing_item_root = listing_item_root
        listing_build.listing_last_update = listing_item_root.last_updated

        # figure out the actual config for this depth
        item_creation_mapping = {}

        def do_list(listing, depth):
            for item in listing.list():
                if depth == 0:
                    yield item
                elif item.is_listable:
                    for sub_item in do_list(item, depth=depth - 1):
                        yield sub_item

        for item in do_list(listing, config["level"]["listing_depth"]):
            if not item.is_listable and config["edge_type"] == "folder":
                continue

            if config["edge_type"] == "file" and not item.is_streamable:
                continue

            item_path = item.path[len(listing.id) :]
            item_path = "%s/%s" % (path.strip("/"), item_path.strip("/"))
            item_creation_mapping[item_path] = item

        listing_build.item_creation_mapping = item_creation_mapping

        # find already existing entries
        existing_listing_items = {}
        for listingitem in ListingItem.objects.filter(
            app=self.service.name, path__in=item_creation_mapping.keys()
        ):
            existing_listing_items[listingitem.path] = listingitem

        logger.info("Found %i existing listingitems" % len(existing_listing_items))
        listing_build.existing_listing_items = existing_listing_items

        indexer_delete_paths = []
        is_update = False
        # delete items that no longer exist at source
        with transaction.atomic():
            if listing_item_root.is_root:
                logger.info("This is an update, looking for old cruft to delete.")
                is_update = True
                for listingitem in ListingItem.objects.filter(
                    app=self.service.name, parent=listing_item_root
                ):
                    if listingitem.path not in existing_listing_items:
                        indexer_delete_paths.append(listingitem.path)
                        listingitem.delete()

            listing_item_root.datetime = listing_item_root.datetime or now()
            listing_item_root.is_root = True
            listing_item_root.last_checked = now()
            listing_item_root.save()

            item_creation_mapping[listing_item_root.path] = listing
            existing_listing_items[listing_item_root.path] = listing_item_root

        # prepare the text search indexer
        indexer = config["level"].get("indexer")
        if indexer:
            if not is_update:
                indexer.clear(path)
            index_writer = indexer.get_writer(path)
            listing_build.index_writer = index_writer

            for indexer_delete_path in indexer_delete_paths:
                index_writer.delete(indexer_delete_path)
        else:
            index_writer = None

        del indexer_delete_paths

        # find listingitems to create and modify existing
        listingitem_children_purges = []
        listingitems = []
        for item_path, item in item_creation_mapping.items():
            if item_path in existing_listing_items:
                listingitem = existing_listing_items[item_path]
                should_create = False
            else:
                listingitem = ListingItem(app=self.service.name, path=item_path)
                listingitem.name = listingitem.get_name()
                should_create = True

            if listingitem != listing_item_root:
                listingitem.parent = listing_item_root

            if (
                listingitem.datetime != item.modified
                and listingitem.is_root
                and listing_item_root != listingitem
            ):
                listingitem.is_root = False
                listingitem_children_purges.append(listingitem)

            listingitem.datetime = item.modified

            listingitem.last_updated = item.modified
            if listingitem.is_root:
                listingitem.item_type = "folder"
            else:
                listingitem.item_type = config["edge_type"]

                # write item name to index if index exists
                if index_writer:
                    normalized_name = (
                        item.id.replace("_", " ").replace(".", " ").replace("-", " ")
                    )
                    index_writer.index(item_path, normalized_name)

            listingitem.attributes = {"name": item.id}
            listingitem.config = {
                "original_item": item.serialize(
                    include_routes=True, include_nested=False
                )
            }

            if "metadata" in item:
                listingitem.attributes["metadata"] = item["metadata"]

            if "size" in item:
                listingitem.attributes["size"] = item["size"]

            if should_create:
                listingitems.append(listingitem)

            logger.trace(
                "Creating:%s listingitem with path:%s and parent path:%s"
                % (should_create, listingitem.path, listing_item_root.path)
            )

        # save the newly created listingitems
        with transaction.atomic():
            for listingitem in existing_listing_items.values():
                listingitem.save()

            for listingitem in listingitem_children_purges:
                ListingItem.objects.filter(
                    app=self.service.name, parent=listingitem
                ).delete()

        ListingItem.objects.bulk_create(listingitems)

        logger.info("Done creating listingitems, linking with metadata")

        self.link_with_metadata(listing_build)

        if index_writer:
            index_writer.commit()

        return listing_item_root

    def link_with_metadata(self, listing_build):
        """
        Link listingitems with metadata
        """
        logger.debug(
            f"Linking listingitems with metadata from root:{listing_build.listing_item_root.path}"
        )

        listing_build.listingitem_mapping = {
            li.path: li
            for li in ListingItem.objects.filter(
                app=self.service.name, parent=listing_build.listing_item_root
            )
        }
        metadata_handlers = sorted(
            listing_build.config["level"]["metadata_handlers"],
            key=lambda x: getattr(x, "priority", 0),
            reverse=True,
        )
        listing_build.linkable_metadata = [
            mh for mh in metadata_handlers if mh.linkable
        ]
        for metadata_handler in metadata_handlers:
            logger.debug(
                f"Linking items from root {listing_build.listing_item_root.path} with metadata handler {metadata_handler.name}"
            )
            metadata_handler.link_metadata_listingitems(
                listing_build, listing_build.config.get("fetch_metadata", True)
            )



import logging
import uuid

from rest_framework import serializers

from ...bases.metadata.mixins import CoverSerializerMixin, PopulateMetadataJSONAPIMixin
from ...bases.metadata.schema import MetadataSchema
from ...plugins import MetadataHandlerPlugin
from .filters import EmbeddedInfoFilter
from .models import EmbeddedInfo, ListingItemRelation

logger = logging.getLogger(__name__)


class EmbeddedInfoSerializer(CoverSerializerMixin, serializers.ModelSerializer):
    id = serializers.SerializerMethodField("get_metadata_identifier")
    type = serializers.SerializerMethodField("get_metadata_type")
    populated = serializers.SerializerMethodField()
    genres = serializers.SerializerMethodField()

    cover = serializers.SerializerMethodField("get_cover_url")

    class Meta:
        model = EmbeddedInfo
        exclude = ("query_key", "path")

    def get_metadata_identifier(self, obj):
        return obj.identifier

    def get_metadata_type(self, obj):
        return "metadata_%s" % obj.metadata_name

    def get_populated(self, obj):
        return True

    def get_genres(self, obj):
        return obj.genres


class EmbeddedMetadataHandlerPlugin(
    PopulateMetadataJSONAPIMixin, MetadataHandlerPlugin
):
    plugin_name = "embedded"
    priority = -10
    config_schema = MetadataSchema
    linkable = False

    model = EmbeddedInfo
    filter = EmbeddedInfoFilter
    listing_item_relation_model = ListingItemRelation
    select_related = []
    prefetch_related = []
    serializer = EmbeddedInfoSerializer
    metadata_embed_method = "relate"  # choices: include, relate, embed

    data_mapping = {
        "metadata": {
            "title": "title",
            "cover": "cover",
            "year": "year",
            "source": "file_source",
            "plot": "plot",
            "genres": "genres",
            "primary_language": "primary_language",
            "rating": "rating",
            "duration": "duration",
            "available": "available",
            "group": "release_group",
        },
        "metadata:bittorrent": {
            "seeders": "bittorrent_seeders",
            "leechers": "bittorrent_leechers",
            "snatched": "bittorrent_snatched",
        },
        "metadata:mediainfo": {
            "resolution": "mediainfo_resolution",
            "codec": "mediainfo_codec",
            "container": "mediainfo_container",
            "source": "mediainfo_source",
            "scene": "mediainfo_scene",
            "audio": "mediainfo_audio",
            "dual_audio": "mediainfo_dual_audio",
            "best": "mediainfo_best",
        },
        "metadata:episodeinfo": {
            "season": "episodeinfo_season",
            "episode": "episodeinfo_episode",
            "year": "episodeinfo_year",
            "month": "episodeinfo_month",
            "day": "episodeinfo_day",
            "episode_type": "episodeinfo_episode_type",
            "sub_title": "episodeinfo_sub_title",  # with a space to avoid confusing with subtitles... hopefully
        },
    }

    def get_metadata(self, request, identifier):
        """
        Not something this type of metadata can actually handle
        """
        pass

    def link_metadata_listingitems(self, listing_build, fetch_metadata=True):
        item_creation_mapping = listing_build.item_creation_mapping
        listingitem_mapping = listing_build.listingitem_mapping

        query_key = str(uuid.uuid4())
        ListingItemRelation.objects.filter(
            listingitem__in=listingitem_mapping.values()
        ).delete()
        objs = []
        for path, item in item_creation_mapping.items():
            if path not in listingitem_mapping:
                continue

            obj = self.model(path=path, query_key=query_key)

            for key, values in self.data_mapping.items():
                if key not in item:
                    continue

                data = item[key]
                if not isinstance(data, dict):
                    continue

                for source_key, target in values.items():
                    if source_key not in data:
                        continue
                    setattr(obj, target, data[source_key])

            if "metadata:index" in item:
                obj.index = item["metadata:index"]

            objs.append(obj)

        self.model.objects.bulk_create(objs)
        lirs = []
        for obj in self.model.objects.filter(query_key=query_key):
            lirs.append(
                self.listing_item_relation_model(
                    metadata=obj, listingitem=listingitem_mapping[obj.path]
                )
            )
        self.listing_item_relation_model.objects.bulk_create(lirs)



import logging

from django.apps import AppConfig as DjangoAppConfig

logger = logging.getLogger(__name__)


class AppConfig(DjangoAppConfig):
    name = "tridentstream.services.imgcache"
    verbose_name = "ImgCache Service"
    label = "services_imgcache"

    def ready(self):
        try:
            from .handler import ImageCacheServicePlugin  # NOQA
            from .imgcache import ImgCachePlugin  # NOQA
        except ImportError:
            logger.exception("Unable to load image cache plugin, please install pillow")



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.metadata.thetvdb"
    verbose_name = "The TVDB Metadata"
    label = "metadata_thetvdb"

    def ready(self):
        try:
            from .handler import TheTVDBMetadataHandlerPlugin  # NOQA
        except ImportError:
            pass



from unplugged import RelatedPluginField, Schema, ServicePlugin

from ...plugins import ImageCachePlugin


class MetadataSchema(Schema):
    image_cache = RelatedPluginField(plugin_type=ImageCachePlugin)
    metadata_server = RelatedPluginField(
        plugin_type=ServicePlugin, traits=["metadata_server"]
    )



from unplugged import Schema

from ...plugins import DatabasePlugin


class MemoryDatabasePlugin(dict, DatabasePlugin):
    plugin_name = "memory"

    config_schema = Schema

    def __init__(self, config):
        pass

    def unload(self):
        self.clear()

    def sync(self):
        pass

    def close(self):
        pass



# Generated by Django 2.2.8 on 2019-12-13 16:29

import django.db.models.deletion
import jsonfield.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("unplugged", "0008_auto_20190715_1944"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("services_listing", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Metadata",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("priority", models.SmallIntegerField(default=20)),
                (
                    "last_update_status",
                    models.CharField(
                        choices=[
                            ("do-not-fetch", "Do not fetch"),
                            ("pending", "Pending"),
                            ("updating", "Updating"),
                            ("failed", "Failed"),
                            ("success", "Success"),
                        ],
                        db_index=True,
                        default="pending",
                        max_length=20,
                    ),
                ),
                ("last_updated", models.DateTimeField(null=True)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("identifier", models.CharField(max_length=255, unique=True)),
                ("populated", models.BooleanField(default=False)),
                ("title", models.CharField(max_length=300, null=True)),
                ("cover", models.URLField(max_length=1000, null=True)),
                (
                    "prefetch_related_denormalized",
                    jsonfield.fields.JSONField(blank=True, null=True),
                ),
                ("name", models.CharField(max_length=200)),
                ("year", models.IntegerField(blank=True, null=True)),
                ("season", models.IntegerField(blank=True, null=True)),
                ("episode", models.IntegerField(blank=True, null=True)),
                (
                    "episode_type",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                ("source", models.CharField(blank=True, max_length=100, null=True)),
                ("fix_type", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "plugin",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata_metadata_iteminfo",
                        to="unplugged.Plugin",
                    ),
                ),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="ListingItemRelation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "listingitem",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata_iteminfo",
                        to="services_listing.ListingItem",
                    ),
                ),
                (
                    "metadata",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="metadata_iteminfo.Metadata",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata_iteminfo",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
                "unique_together": {("listingitem", "metadata", "user")},
            },
        ),
    ]



import rest_framework_filters as filters

from .models import ListingItemRelation


class EmbeddedInfoFilter(filters.FilterSet):
    index = filters.DateTimeFilter(field_name="metadata__index", name="metadata__index")
    season = filters.AllLookupsFilter(
        field_name="metadata__episodeinfo_season", name="metadata__episodeinfo_season"
    )
    episode = filters.AllLookupsFilter(
        field_name="metadata__episodeinfo_episode", name="metadata__episodeinfo_episode"
    )

    class Meta:
        model = ListingItemRelation
        order_by = ["index", "season", "episode"]
        fields = []



import logging
from datetime import datetime

import pytz
from django.http import Http404, HttpResponseNotModified
from django.utils.http import parse_http_date_safe
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView
from thomas import Item
from unplugged import CascadingPermission

from ...exceptions import NotModifiedException
from ...plugins import InputPluginManager
from ...stream import create_stream

logger = logging.getLogger(__name__)


class RemoteFilesystemView(APIView):
    service = None

    permission_classes = (CascadingPermission,)

    def get_plugin_path(self, path):
        path += "/"
        input_name, path = path.split("/", 1)

        for input_config in self.service.config["inputs"]:
            if input_config["input"].name != input_name:
                continue

            return input_config["input"], path.strip("/")

        return None, None

    def get(self, request, path):
        path = path.strip("/")
        if path == "":
            listing = Item(id="")
            for input_config in self.service.config.get("inputs", []):
                if not input_config or not input_config["input"]:
                    continue
                plugin = input_config["input"]

                item = plugin.get_item("")
                item.id = plugin.name
                listing.add_item(item)

            return Response(listing.serialize())
        else:
            if_modified_since = request.META.get("HTTP_IF_MODIFIED_SINCE")
            if if_modified_since:
                if_modified_since = parse_http_date_safe(if_modified_since)
            if if_modified_since:
                if_modified_since = datetime.fromtimestamp(if_modified_since, pytz.UTC)

            depth = int(request.GET.get("depth", 0))

            plugin, plugin_path = self.get_plugin_path(path)

            if not plugin:
                raise Http404

            item = InputPluginManager.get_item(plugin, plugin_path)

            logger.info(
                f"Trying to create listing for {plugin!r} - path:{plugin_path} - last_modified:{if_modified_since}"
            )
            item.list(depth=depth)

            if not item.is_readable and not item.is_listable:
                raise Http404

            if if_modified_since and item.modified <= if_modified_since:
                return HttpResponseNotModified()

            return Response(item.serialize())

    def post(self, request, path):
        plugin, plugin_path = self.get_plugin_path(path)
        if not plugin:
            raise Http404

        item = InputPluginManager.get_item(plugin, plugin_path)

        try:
            streamresult = create_stream(item, request)
            return Response(streamresult.serialize(request))
        except:
            logger.exception(f"Failed to stream path:{path!r} from {plugin.name}")

        return Response(
            {"status": "failed", "message": "Unable to properly stream "},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )



from abc import abstractmethod

from unplugged import PluginBase


class ItemInfoPlugin(PluginBase):
    plugin_type = "iteminfo"

    @abstractmethod
    def get_info(self, name):
        """
        Returns a dictionary with information about the name
        """



from abc import abstractmethod

from unplugged import PluginBase


class BittorrentClientPlugin(PluginBase):
    plugin_type = "bittorrentclient"

    @abstractmethod
    def stream(self, info_hash, torrent_filelike_data, path=None):
        """
        Return a stream object to the file with given infohash,
        said torrent data and given file_path inside torrent.

        torrent_filelike_data must be a filelike object.
        empty file_path makes the client decide what to stream (most likely the biggest file).
        """



import rest_framework_filters as filters

from .models import ListingItemRelation


class MetadataFilter(filters.FilterSet):
    tag_name = filters.AllLookupsFilter(
        field_name="metadata__tag_name", name="metadata__tag_name"
    )
    created = filters.AllLookupsFilter(
        field_name="metadata__created", name="metadata__created"
    )

    class Meta:
        model = ListingItemRelation
        fields = []
        include_related = []
        order_by = ["created"]
        user_field = "metadata__user"



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.notifiers.wamp"
    verbose_name = "WAMP Notifier"
    label = "notifier_wamp"

    def ready(self):
        from .handler import WAMPNotifierHandlerPlugin  # NOQA



import logging

from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from rest_framework import serializers

from ...bases.metadata.linkingmetadata import LinkingMetadataHandlerPlugin
from .filters import MetadataFilter
from .models import FirstSeen, ListingItemRelation

logger = logging.getLogger(__name__)


class FirstSeenSerializer(serializers.ModelSerializer):
    id = serializers.SerializerMethodField("get_metadata_identifier")
    type = serializers.SerializerMethodField("get_metadata_type")

    class Meta:
        model = FirstSeen
        fields = ("id", "type", "datetime")

    def get_metadata_identifier(self, obj):
        return obj.metadata.identifier

    def get_metadata_type(self, obj):
        return f"metadata_{obj.metadata.metadata_name}"


class FirstSeenMetadataHandlerPlugin(LinkingMetadataHandlerPlugin):
    plugin_name = "firstseen"
    priority = -10

    prefetch_related = ["metadata"]
    serializer = FirstSeenSerializer
    model = FirstSeen
    filter = MetadataFilter
    listing_item_relation_model = ListingItemRelation
    metadata_link_model = FirstSeen
    metadata_embed_method = "relate"

    def link_metadata_listingitems(self, listing_build, fetch_metadata=True):
        listing_item_root = listing_build.listing_item_root
        linkable_metadata = listing_build.linkable_metadata

        logger.info(f"Linking with metadata:{linkable_metadata!r}")
        for p in linkable_metadata:
            content_type = ContentType.objects.get_for_model(p.model)

            base_qs = p.listing_item_relation_model.objects.filter(
                Q(listingitem__parent=listing_item_root)
                | Q(listingitem=listing_item_root)
            )
            existing_items = self.metadata_link_model.objects.filter(
                content_type=content_type, object_id__in=base_qs.values("metadata_id")
            )
            missing_items = base_qs.exclude(
                metadata_id__in=existing_items.values("object_id")
            )

            create_objs = []
            for metadata_id in missing_items.values_list(
                "metadata_id", flat=True
            ).distinct():
                logger.debug(
                    f"Preparing to create object_id:{metadata_id!r}, content_type:{content_type!r}"
                )
                create_objs.append(
                    self.model(object_id=metadata_id, content_type=content_type)
                )
            self.model.objects.bulk_create(create_objs)
            logger.debug(f"Created {len(create_objs)} missing objects")

        return super(FirstSeenMetadataHandlerPlugin, self).link_metadata_listingitems(
            listing_build, fetch_metadata
        )

    def get_metadata(self, request, identifier):
        pass



import logging

from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream"
    verbose_name = "Tridentstream"

    def ready(self):
        from .eventtrack import register_signals

        register_signals()



import logging
from datetime import timedelta

from django.contrib.contenttypes.models import ContentType
from django.db.models import Count
from django.utils.timezone import now
from unplugged import RelatedPluginField, Schema

from ...plugins import HistoryPlugin, MetadataHandlerPlugin

logger = logging.getLogger(__name__)

HISTORY_WATCH_PERIOD = timedelta(hours=24)


class HistorySchema(Schema):
    history_metadata = RelatedPluginField(
        plugin_type=MetadataHandlerPlugin, traits=["history"]
    )


class HistoryHistoryPlugin(HistoryPlugin):
    plugin_name = "history"
    config_schema = HistorySchema

    __traits__ = ["queryable"]

    def __init__(self, config):
        self.history = config["history_metadata"]

    def log_history(self, config, listingitem, viewstate):
        level_config = config.get("level", {})
        histories = self.history.model.objects.filter(
            user=viewstate.user,
            name=listingitem.name,
            last_watched__gte=now() - HISTORY_WATCH_PERIOD,
        )
        logger.debug(
            f"Trying to log history entry with listingitem:{listingitem!r} to user:{viewstate.user!r}"
        )

        if histories:
            history = histories[0]
            logger.debug(f"Reusing old history entry:{history!r}")
        else:
            history = self.history.model(user=viewstate.user, name=listingitem.name)
            logger.debug("Creating new history entry")

        history.listingitem_app = listingitem.app
        history.listingitem_path = listingitem.path
        history.last_watched = now()
        history.save()

        for metadata_handler in level_config.get("metadata_handlers", []):
            if not metadata_handler.linkable:
                continue

            primary_metadata = "primary_metadata" in metadata_handler.__traits__

            for metadata in listingitem.get_metadata(metadata_handler.plugin_name):
                content_type = ContentType.objects.get_for_model(metadata)

                try:
                    metadata_link = self.history.metadata_link_model.objects.get(
                        history=history,
                        object_id=metadata.pk,
                        content_type=content_type,
                    )
                except self.history.metadata_link_model.DoesNotExist:
                    metadata_link = self.history.metadata_link_model.objects.create(
                        history=history,
                        metadata=metadata,
                        primary_metadata=primary_metadata,
                    )

                self.history.listing_item_relation_model.objects.get_or_create(
                    metadata=metadata_link, listingitem=listingitem, user=viewstate.user
                )

        try:
            viewstate_obj = self.history.viewstate_model.objects.get(
                identifier=viewstate.id, user=viewstate.user
            )
        except self.history.viewstate_model.DoesNotExist:
            viewstate_obj = self.history.viewstate_model(
                identifier=viewstate.id, user=viewstate.user, values={}
            )

        viewstate_obj.history = history
        viewstate_obj.save()

        if viewstate_obj.values:
            logger.debug(
                f"Existing viewstate already there for id {viewstate_obj.identifier}, updating with new values"
            )
            viewstate.update(viewstate_obj.values)

        viewstate.add_change_callback(self.viewstate_updated)

    def viewstate_updated(self, viewstate):
        try:
            viewstate_obj = self.history.viewstate_model.objects.get(
                identifier=viewstate.id, user=viewstate.user
            )
        except self.history.viewstate_model.DoesNotExist:
            logger.warning(
                f"Got update to unknown viewstate {viewstate.id} for user {viewstate.user}"
            )
            return

        viewstate_obj.values.update(viewstate)
        viewstate_obj.save()
        logger.debug(
            f"Updated viewstate {viewstate.id} for user {viewstate.user} with {viewstate_obj.values}"
        )



import logging

from unplugged import Schema, fields
from whoosh import analysis, index
from whoosh.fields import ID, TEXT, SchemaClass
from whoosh.qparser import QueryParser
from whoosh.query import Term
from whoosh.writing import AsyncWriter

from ...plugins import IndexerPlugin, PathIndexer

logger = logging.getLogger(__name__)


class WhooshSchema(SchemaClass):
    parent_path = ID()
    path = ID(stored=True)
    title = TEXT(
        spelling=False, analyzer=analysis.StandardAnalyzer(minsize=1, stoplist=[])
    )


class WhooshIndexerSchema(Schema):
    path = fields.String()


class WhooshPathIndexer(PathIndexer):
    def __init__(self, writer, parent_path):
        self.writer = writer
        self.parent_path = parent_path

    def index(self, path, title):
        logger.trace(
            f"Indexing parent_path:{self.parent_path}, path:{path}, title:{title}"
        )

        self.writer.add_document(parent_path=self.parent_path, path=path, title=title)

    def delete(self, path):
        logger.trace(f"Deleting all indexes for path {path}")
        self.writer.delete_by_query(Term("path", path))

    def commit(self):
        logger.debug(f"Committing {self.parent_path}")
        self.writer.commit()


class WhooshIndexerPlugin(IndexerPlugin):
    plugin_name = "whoosh"
    config_schema = WhooshIndexerSchema

    def __init__(self, config):
        path = self.get_database_path(config.get("path"))

        try:
            self.ix = index.open_dir(path)
        except index.EmptyIndexError:
            self.ix = index.create_in(path, WhooshSchema())

    def clear(self, parent_path):
        logger.debug(f"Clearing index for {self.name}")
        self.ix.delete_by_query(Term("parent_path", parent_path))

    def get_writer(self, parent_path):
        logger.debug(f"Getting index writer for path:{parent_path}")
        return WhooshPathIndexer(AsyncWriter(self.ix), parent_path)

    def search(self, parent_path, query):
        qp = QueryParser("title", schema=self.ix.schema)
        q = qp.parse(query) & Term("parent_path", parent_path)

        with self.ix.searcher() as s:
            results, seen = [], set()
            for path in [x["path"] for x in s.search(q, limit=120)]:
                if path in seen:
                    continue

                seen.add(path)
                results.append(path)

        logger.debug(f"Searched for {query!r} found {len(results)} results")
        return results

    def unload(self):
        logger.debug(f"Unloading index {self.name}")
        self.ix.close()



import logging

import requests
from django.core.cache import cache
from thomas import Item
from unplugged import Schema, fields

from ...bases.websearcher.torrentsearcher import URLSearcherPluginBase
from ...exceptions import PathNotFoundException
from ...plugins import SearcherFilter, SearcherPlugin
from ...stream import Stream
from ...utils import urljoin

logger = logging.getLogger(__name__)


class RemoteStreamer:
    def __init__(self, plugin, stream_info, sub_path, priority):
        self.plugin = plugin
        self.stream_info = stream_info
        self.sub_path = sub_path
        self.priority = priority

    def evaluate(self):
        return self.priority

    def stream(self):
        return self.plugin.stream(self.stream_info, self.sub_path)


class RemoteSearcherSearcherSchema(Schema):
    url = fields.String()
    token = fields.String()
    searcher_name = fields.String()
    priority = fields.Integer(default=10)


class RemoteSearcherSearcher(URLSearcherPluginBase):
    plugin_name = "remotesearcher"
    config_schema = RemoteSearcherSearcherSchema

    simpleadmin_templates = True

    base_url = None
    search_path = ""

    @property
    def base_url(self):
        return urljoin(self.config["url"], self.config["searcher_name"])

    def get_headers(self):
        return {"Authorization": f"Token {self.config['token']}"}

    def do_query(self, session, url, query):
        return session.get(
            self.build_url(url, query), headers=self.get_headers()
        ).json()

    def stream(self, stream_info, sub_path):
        url = urljoin(self.base_url, stream_info["search_token"], stream_info["path"])
        params = {"sub_path": sub_path}
        with self.get_session() as session:
            r = session.post(url, headers=self.get_headers(), params=params)

        if r.status_code != 200:
            raise PathNotFoundException()

        return Stream.unserialize(r.json())

    def thomas_stream(self, item, stream_info, sub_path):
        return RemoteStreamer(self, stream_info, sub_path, self.config["priority"])

    def add_expand_routes(self, item, search_token, url, path="", skip=False):
        if not skip:
            if path:
                path = f"{path}/{item.id}"
            else:
                path = item.id

        if item.is_expanded:
            for nested_item in item.nested_items:
                self.add_expand_routes(nested_item, search_token, url, path)
        elif item.expandable:
            self.make_expandable(
                item, search_token, url, {"sub_path": path}, path, None
            )

    def parse_result(self, search_token, session, response, response_type):
        item = Item.unserialize(response["item"])
        self.add_stream_routes(
            item,
            {"search_token": response["search_token"], "path": response["path"]},
            ensure_already_streamable=True,
        )
        url = urljoin(self.base_url, response["search_token"], response["path"])
        self.add_expand_routes(item, search_token, url)
        return item

    @property
    def field_rewrite(self):
        return {k: k for k in self.filters.fields}

    @property
    def filters(self):
        cache_key = f"searcher:filters:{self.name}"
        cache_result = cache.get(cache_key)
        if not cache_result:
            r = requests.get(self.config["url"], headers=self.get_headers())
            if r.status_code != 200:
                return None
            r = r.json()
            cache.set(cache_key, r, 15 * 60)
        else:
            r = cache_result

        for f in r:
            if f["name"] == self.config["searcher_name"]:
                return SearcherFilter.unserialize(f["filters"])

    def check_site_usable(self, session):
        return True



import io
import logging
from urllib.parse import quote_plus

import requests
from django.core.cache import cache
from django.core.files.storage import default_storage
from django.core.signing import Signer
from unplugged import RelatedPluginField, Schema, ServicePlugin

from ...plugins import ImageCachePlugin
from ...utils import hash_string

logger = logging.getLogger(__name__)

ALLOWED_FILE_EXTENSIONS = {"png", "jpg", "jpeg", "gif"}


class ImgCachePluginSchema(Schema):
    image_server = RelatedPluginField(plugin_type=ServicePlugin, traits=["image_cache"])


class ImgCachePlugin(ImageCachePlugin):
    plugin_name = "imgcache"

    config_schema = ImgCachePluginSchema

    def get_image_path(self, url):
        if cache.get(f"imgcache:{self.name}:{url}") == "F":
            return None

        ext = url.split("?")[0].split(".")[-1].lower()
        if ext not in ALLOWED_FILE_EXTENSIONS:
            logger.warning(f"Unknown file-extension: {ext}")
            return None

        filename = f"{hash_string(url)}.{ext}"
        path = f"{self.plugin_name}_{self.name}/{filename}"

        if not default_storage.exists(path):
            request_failed = False
            try:
                r = requests.get(url, verify=False, timeout=8)
            except requests.exceptions.RequestException:
                logger.warning("Requests errored out")
                request_failed = True
            else:
                if r.status_code != 200:
                    request_failed = True

            if request_failed:
                cache.set(f"imgcache:{self.name}:{url}", "F", 30 * 60)
                logger.warning(f"Failed to fetch url {url}")
                return None

            logger.debug(f"Saving {url} to {path}")

            default_storage.save(path, io.BytesIO(r.content))

        return default_storage.path(path)

    def get_image_url(self, request, url):
        signer = Signer()
        signed_url = signer.sign(url)
        result_path = f"/{self.config['image_server'].name}/{self.name}?url={quote_plus(signed_url)}"

        return request.build_absolute_uri(result_path)

    def unload(self):
        pass
